""" This file contains the draw functions for align faces operator addon"""
import math
import os
import mathutils
import bpy
import blf
import bgl
import gpu
from gpu_extras.batch import batch_for_shader

# Import From Files
if "bpy" in locals():
    import importlib
    from . import shaders
    importlib.reload(shaders)
    from . import mesh
    importlib.reload(mesh)
    from . import states
    importlib.reload(states)
    from . import helpers
    importlib.reload(helpers)
    from .helpers import addon_prefs
    from . config import settings
    importlib.reload(settings)
else:
    from . import states
    from . import shaders
    from . import mesh
    from . import helpers
    from .helpers import addon_prefs
    from . config import settings

ADJUSTLINES = 3
LINESPACING = 5


def get_ortho(a_pos, b_pos, c_pos):
    """ calculate a perpendicular vector to a given one """
    if c_pos != 0.0 and -a_pos != b_pos:
        return [-b_pos-c_pos, a_pos, a_pos]
    return [c_pos, c_pos, -a_pos-b_pos]


def draw_circle_2d(color, cx, cy, r, num_segments, thickness):
    theta = 2 * 3.1415926 / num_segments
    c = math.cos(theta)  # precalculate the sine and cosine
    s = math.sin(theta)
    x = r  # we start at angle = 0
    y = 0

    coords = []
    for i in range(num_segments):
        # bgl.glVertex2f(x + cx, y + cy) # output vertex
        coords.append((x + cx, y + cy, 0))
        # apply the rotation matrix
        t = x
        x = c * x - s * y
        y = s * t + c * y

    draw_line(color, coords, thickness)


def draw_box_2d(color, sx, sy, ex, ey, num_segments, thickness):

    coords = []
    coords = [(sx, sy, 0), (ex, sy, 0),  (ex, sy, 0), (ex, ey, 0),
               (ex, ey, 0), (sx, ey, 0), (sx, ey, 0), (sx, sy, 0)]
    """ for i in range (num_segments):
        # bgl.glVertex2f(x + cx, y + cy) # output vertex
        coords.append((x + cx, y + cy, 0))
        # apply the rotation matrix
        t = x
        x = c * x - s * y
        y = s * t + c * y """

    draw_line(color, coords, thickness)


def draw_arrow(color_up, coords, thickness):
    """ draw a two colored line with bgl """
    draw_line(color_up, coords, thickness)

    # coords[0] = get_ortho(coords[1][0], coords[1][1], 0)
    coords = [get_ortho(coords[1][0], coords[1][1], 0),
              (coords[1][0], coords[1][1]-thickness, 0)]
    draw_line(color_up, coords, thickness)
    # draw_aline_r(color_down, coords, thickness)


def draw_double_line(color_up, color_down, coords, thickness):
    """ draw a two colored line with bgl """
    draw_line(color_up, coords, thickness)
    coords = [(coords[0][0], coords[0][1]-thickness, 0),
              (coords[1][0], coords[1][1]-thickness, 0)]
    draw_line(color_down, coords, thickness)


def draw_line(color, coords, thickness):
    """ draw a colored line with bgl i a given thickness """
    # shader = gpu.shader.from_builtin('2D_UNIFORM_COLOR')
    shader = gpu.shader.from_builtin('3D_UNIFORM_COLOR')
    batch = batch_for_shader(shader, 'LINES', {"pos": coords})

    bgl.glEnable(bgl.GL_BLEND)

    bgl.glLineWidth(thickness)
    shader.bind()
    shader.uniform_float("color", color)
    batch.draw(shader)
    bgl.glLineWidth(1)  # restore opengl defaults

    bgl.glDisable(bgl.GL_BLEND)


def draw_text(font_id, text, color, coords):
    """ here we draw some text via bgl """
    blf.color(font_id, color[0], color[1], color[2], color[3])
    blf.shadow(font_id, 3, 0.0, 0.0, 0.0, 0.75)
    blf.shadow_offset(font_id, 2, -3)
    blf.enable(font_id, 4)
    blf.position(font_id, coords[0], coords[1], 0)
    blf.draw(font_id, text)


# a debug line is 3-tupple containing a coords tuple, a color and thickness
debug_lines = []
debug_polys = []


def draw_outline(obj):
    pass


debug_objects = []

rotflip_objects = set()
wireframe_rotflip_batches = []
wireframe_batch = None
rotflip_needs_to_be_recreated = False

def create_wireframe_batch(context, coords, matrix=None, shader=shaders.rgb_3d, line_type='LINES'):
    """ draw a lines with bgl. Coordinates are given as pairs of start -and endpoint"""
    global wireframe_batch
    wireframe_batch = batch_for_shader(shader, line_type, {"position": coords})

def draw_wireframe_batch(context,  color, thickness, matrix=None, shader=shaders.rgb_3d):
    """ draw a lines with bgl. Coordinates are given as pairs of start -and endpoint"""
    global wireframe_batch
    # if no matrix is given assume coordinates are in world space and use perspective matrix only
    if matrix is None:
        matrix = context.region_data.perspective_matrix

    bgl.glEnable(bgl.GL_BLEND)

    bgl.glLineWidth(thickness)

    shader.bind()
    shader.uniform_float("MVP_Matrix", matrix)
    shader.uniform_float("color", color)
    wireframe_batch.draw(shader)

    bgl.glLineWidth(1)  # restore opengl defaults

    bgl.glDisable(bgl.GL_BLEND)


def update_rotflip_objects(selected_objects, context):
    """ checks if a change in the set of selected rotation objects occured and updates
    then the rotflip_objects """
    global rotflip_needs_to_be_recreated
    global rotflip_objects
    context.window_manager.f2f_tmp_vars.update_rotflip_objects = False
    new_rotflip_objects = set()
    for sel_obj in selected_objects:
        if sel_obj.rotation_flipped:
            new_rotflip_objects.add(sel_obj)

    if rotflip_objects == new_rotflip_objects:

        rotflip_needs_to_be_recreated = False
        return False
    
    rotflip_objects = new_rotflip_objects

    rotflip_needs_to_be_recreated = True
    return True



def create_wireframe_rotflip_batches(shader=shaders.rgb_3dr, line_type='LINES'):
    """ draw a lines with bgl. Coordinates are given as pairs of start -and endpoint"""
    global wireframe_rotflip_batches
    global rotflip_objects
    global rotflip_needs_to_be_recreated
    wireframe_rotflip_batches = []
    
    for obj in rotflip_objects:
        coords = []
        for edge in obj.data.edges:
            verts = edge.vertices
            #v1_wc = obj.matrix_local @ obj.data.vertices[verts[0]].co
            #v2_wc = obj.matrix_local @ obj.data.vertices[verts[1]].co
            v1_wc = obj.data.vertices[verts[0]].co
            v2_wc = obj.data.vertices[verts[1]].co
            coords.append(v1_wc)
            coords.append(v2_wc)
        wireframe_rotflip_batches.append(( batch_for_shader(shader, line_type, {"position": coords}), obj))
    
    rotflip_needs_to_be_recreated = False

def draw_wireframe_rotflip_batches(context, color, thickness, shader=shaders.rgb_3dr):
    """ draw a lines with bgl. Coordinates are given as pairs of start -and endpoint"""
    global wireframe_rotflip_batches
    #if matrix is None:
    matrix_p = context.region_data.perspective_matrix
    #matrix_v = context.region_data.view_matrix
    #print("draw rotbatch atrix p:", matrix_p)
    bgl.glEnable(bgl.GL_BLEND)
    bgl.glLineWidth(thickness)
    for rotflip_tuple in wireframe_rotflip_batches:
        # take individual object matrix to reflect changes made during rotation
        

        batch = rotflip_tuple[0] 
        obj = rotflip_tuple[1]
        #print("draw rotbatch:", obj.name)
        #matrix = obj.matrix_local @ matrix_v @ matrix_p
        #matrix = obj.matrix_local.copy()# @ matrix_p.copy()
        #matrix = obj.matrix_local.copy() @ matrix_v.copy() @ matrix_p.copy()
        matrix = matrix_p @ obj.matrix_world
        #print("draw rotbatch atrix:", matrix)

        shader.bind()
        shader.uniform_float("MVP_Matrix", matrix)
        shader.uniform_float("color", color)
        batch.draw(shader)

    bgl.glLineWidth(1)  # restore opengl defaults

    bgl.glDisable(bgl.GL_BLEND)

def draw_wireframe(context, obj, color):
    for edge in obj.data.edges:
        verts = edge.vertices
        v1_wc = obj.matrix_local @ obj.data.vertices[verts[0]].co
        v2_wc = obj.matrix_local @ obj.data.vertices[verts[1]].co
        coords = [v1_wc, v2_wc]
        draw_lines_3d(context,
            coords, color, 1)


def highlight_objects(context, modal):
    highlight_non_selected_projected_objects(context, modal)
    highlight_selected_rotflip_objects(context, modal)


def highlight_non_selected_projected_objects(context, modal):
    """ highlight all not selected projected objects
         to visaly showcase our working set """
    global rotflip_needs_to_be_recreated

    if context.window_manager.f2f_tmp_vars.selection_changed:
        context.window_manager.f2f_tmp_vars.selection_changed = False
        coords = []
        for obj in modal.projected_objects:
            if obj not in  modal.selected_objects:
                for edge in obj.data.edges:
                    verts = edge.vertices
                    v1_wc = obj.matrix_local @ obj.data.vertices[verts[0]].co
                    v2_wc = obj.matrix_local @ obj.data.vertices[verts[1]].co
                    coords.append(v1_wc)
                    coords.append(v2_wc)
        create_wireframe_batch(context, coords)
    if wireframe_batch:
         draw_wireframe_batch(context, context.window_manager.face2face.highlight_projected_objects_color, 1)

def highlight_selected_rotflip_objects(context, modal):
    if isinstance(modal.state, states.RotateState):

        if rotflip_needs_to_be_recreated:
            create_wireframe_rotflip_batches()
        if len(wireframe_rotflip_batches) > 0:

            draw_wireframe_rotflip_batches(context, context.window_manager.face2face.highlight_rotation_flipped_color,1)


def draw_wireframeole(context, obj, color):
    for edge in obj.data.edges:
        verts = edge.vertices
        v1_wc = obj.matrix_local @ obj.data.vertices[verts[0]].co
        v2_wc = obj.matrix_local @ obj.data.vertices[verts[1]].co
        coords = [v1_wc, v2_wc]
        draw_lines_3d(context,
            coords, color, 1)


def highlight_non_selected_projected_objectsole(context, modal):
    """ highlight all not selected projected objects
         to visaly showcase our working set """
    for obj in modal.projected_objects:
        if obj not in  modal.selected_objects:
            draw_wireframe(context, obj, context.window_manager.face2face.highlight_projected_objects_color)

def draw_debug_stuff(context):
    """ draw debug lines and polys set outside of draw """
    for obj in debug_objects:
        draw_wireframe(context, obj, color = (0.27, 0.56, 0.69))

    for line in debug_lines:
        draw_lines_3d(context,
                      line[0], line[1], line[2])

def draw_circle_select(context):
    """ draw debug lines and polys set outside of draw """
    if len(context.window_manager.f2f_tmp_vars.circle_selecting) > 0:
        draw_circle_2d((0.8, 0.8, 0.8, 0.75), context.window_manager.f2f_tmp_vars.circle_selecting[0], context.window_manager.f2f_tmp_vars.circle_selecting[1], context.window_manager.f2f_tmp_vars.circle_selecting[2], 100, 1)


box_selecting = []
def draw_box_select(context):
    """ draw debug lines and polys set outside of draw """
    if len(context.window_manager.f2f_tmp_vars.box_selecting) > 0:
        draw_box_2d((0.8, 0.8, 0.8, 0.75), context.window_manager.f2f_tmp_vars.box_selecting[0], context.window_manager.f2f_tmp_vars.box_selecting[1], context.window_manager.f2f_tmp_vars.box_selecting[2], context.window_manager.f2f_tmp_vars.box_selecting[3],100, 1)



def draw_callback_px(context, modal):
    """ main draw callback funtion from FF modal """
    #global do_cleanup
    if  context.window_manager.face2face.do_cleanup:
        precleanup(context)

    if context.window_manager.f2f_tmp_vars.update_rotflip_objects:
        update_rotflip_objects(context.window_manager.f2f_tmp_vars.rotflip_objects, context)

    draw_debug_stuff(context)
    
    if context.window_manager.face2face.highlight_projected_objects:
        highlight_objects(context, modal)

    draw_circle_select(context)
    draw_box_select(context)

    npanel = next((x for x in context.area.regions if x.type == 'UI'), None)
    npanel_width = npanel.width
    transorm_panel = next(
        (x for x in context.area.regions if x.type == 'TOOLS'), None)
    transorm_panel_width = transorm_panel.width

    width = context.area.width - npanel_width - transorm_panel_width

    state_lines = draw_ui4state(context, modal)

    font_id = 0

    blf.size(font_id, 16, 72)
    dim_a_y = blf.dimensions(font_id, modal.status_text)[1]

    blf.size(font_id, 24, 72)

    dim_x = blf.dimensions(font_id, modal.status_text)[0]/2
    dim_y = blf.dimensions(font_id, modal.status_text)[1]
    
    header_y = (dim_a_y + LINESPACING) * (len(state_lines)+1)
    logo_size = (header_y+dim_y)-2*LINESPACING
    header_x = width/2-dim_x

    if context.window_manager.face2face.show_logo: 
        header_x += (logo_size+LOGO_SPACING)/2

    coords = (header_x, header_y)

    #header_color = context.window_manager.face2face.status_text_error_color if blendout else modal.status_text_header_color
    header_color = modal.status_text_header_color
    draw_text(font_id, modal.status_text, header_color, coords)

    coords = [(header_x, (header_y - LINESPACING), 0),
              (header_x+dim_x*2, (header_y - LINESPACING), 0)]
    draw_double_line((0.8, 0.8, 0.8, 0.75), (0.0, 0.0, 0.0, 0.75), coords, 1)
    
    
    if context.window_manager.face2face.show_logo:
        draw_logo(header_x, header_y+dim_y, logo_size, context)
    # draw  all secondary adjustment lines
    blf.size(font_id, 16, 72)

    for i in range(1, len(state_lines)+1):
        coords = (header_x, (header_y - i*(LINESPACING+dim_a_y) - LINESPACING))
        draw_text(font_id, state_lines[i-1], context.window_manager.face2face.status_text_color, coords)#(1.0, 1.0, 1.0, 0.75)


def draw_selection_ui_text(modal):
    state_lines = []

    projected_text = "Selected  objects : "+str(len(modal.selected_objects))+"/"+str(len(modal.projected_objects))
    state_lines.append(projected_text)

    state_text = "Current state : "+modal.state.name
    state_lines.append(state_text)

    return state_lines

def draw_idle_ui(context, modal):
    pass

def draw_idle_ui_text(modal):
    state_lines = []

    projected_text = "Selected  objects : "+str(len(modal.selected_objects))+"/"+str(len(modal.projected_objects))
    state_lines.append(projected_text)

    projected_text = settings.current_projection_name
    state_lines.append(projected_text)

    return state_lines

def draw_scaling_ui_text(modal):
    state_lines = []

    if modal.state.setting_variation:  # random_scale > 0:
        amount = "MAX" if modal.scale_variation_strength == 1 else str(round(modal.scale_variation_strength,3))
        limit_text = "Random variation strength: "+amount
        state_lines.append(limit_text)
        return state_lines

    if modal.state.setting_relative:  # random_scale > 0:
        amount = "MAX" if modal.scale_relative_strength == 1 else str(round(modal.scale_relative_strength,3))
        limit_text = "Account face size strength: "+amount
        state_lines.append(limit_text)
        return state_lines

    if modal.state.num_enter_mode:  
        limit_text = "Scale precisely by: "+str(modal.state.num_value)+""
        state_lines.append(limit_text)

    pm_text = "NORMAL"  # "RELATIVE" if modal.state.relative else "ABSOLUTE"
    if modal.state.fit_scale:
        pm_text = "FIT"
        if modal.state.even_scale:
            pm_text += " - EVEN"
        else:
            pm_text += " - IMAX"
    state_lines.append("Scaling mode : "+pm_text)
    if modal.state.limit_to_face:
        limit_text = "limit to target face : ON"
        state_lines.append(limit_text)
    else:
        limit_text = "limit to target face : OFF"
        state_lines.append(limit_text)

    if modal.state.axis_locked in (1, 2, 3):
        lock_text = "locked axis: X"
        if modal.state.axis_locked == 2:
            lock_text = "locked axis: Y"
        if modal.state.axis_locked == 3:
            lock_text = "locked axis: Z"
        state_lines.append(lock_text)
    if modal.scale_variation_strength > 0 and not modal.state.fit_scale:  
        amount = "MAX" if modal.scale_variation_strength == 1 else str(round(modal.scale_variation_strength, 3))
        limit_text = "Random variation strength: "+amount
        state_lines.append(limit_text)
    if modal.scale_relative_strength > 0 and not modal.state.fit_scale:  
        amount = "MAX" if modal.scale_relative_strength == 1 else str(round(modal.scale_relative_strength, 3))
        limit_text = "Account face size strength: "+amount
        state_lines.append(limit_text)
    return state_lines


def draw_rotation_ui(context, modal):
    pass

def draw_rotation_ui_text(modal):
    state_lines = []

    if modal.state.setting_variation:  # random_scale > 0:
        amount = "MAX" if modal.rotation_variation_strength == 1 else str(round(modal.rotation_variation_strength, 3))
        limit_text = "Random variation strength: "+amount
        state_lines.append(limit_text)
        return state_lines
    
    if modal.state.num_enter_mode:  
        limit_text = "Setting angle precisely to: "+str(modal.state.num_value)+"°"
        state_lines.append(limit_text)

    projected_text = "Selected  objects : "+str(len(modal.selected_objects))+"/"+str(len(modal.projected_objects))
    state_lines.append(projected_text)

    if modal.rotation_variation_strength > 0:  
        amount = "MAX" if modal.rotation_variation_strength == 1 else str(round(modal.rotation_variation_strength, 3))
        limit_text = "Random variation strength: "+amount
        state_lines.append(limit_text)

    
    
    if len(state_lines) < 2:
        projected_text = settings.current_projection_name
        state_lines.append(projected_text)

    

    return state_lines

def draw_ui4state(context, modal):
    state_lines = []
    if modal.state.select_mode > 0:
        # draw_scaling_ui(context, modal)
        state_lines = draw_selection_ui_text(modal)
        return state_lines

    if isinstance(modal.state, states.ScaleState):
        draw_scaling_ui(context, modal)
        state_lines = draw_scaling_ui_text(modal)
    if isinstance(modal.state, states.RotateState):
        draw_rotation_ui(context, modal)
        state_lines = draw_rotation_ui_text(modal)
    if isinstance(modal.state, states.IdleState):
        draw_idle_ui(context, modal)
        state_lines = draw_idle_ui_text(modal)
    """ if self.next_state == 'BLENDOUT':
        self.state = BlendOutState()
        self.current_state = 'BLENDOUT' """
    if isinstance(modal.state, states.GrabState):
        draw_grabing_ui(context, modal)
        if modal.state.num_enter_mode:  
            limit_text = "Setting offset precisely to: "+str(modal.state.num_value)+""
            state_lines.append(limit_text)
        pm_text = "RELATIVE" if modal.state.relative_pos else "ABSOLUTE"
        state_lines.append("Positioning mode : "+pm_text)
        if modal.state.relative_pos:
            limit_text = "limit to target face : ALWAYS ON"
            state_lines.append(limit_text)
        elif modal.state.limit_to_face:
            limit_text = "limit to target face : ON"
            state_lines.append(limit_text)
        else:
            limit_text = "limit to target face : OFF"
            state_lines.append(limit_text)
        if modal.state.axis_locked in (1, 2, 3):
            lock_text = "locked axis: X"
            if modal.state.axis_locked == 2:
                lock_text = "locked axis: Y"
            if modal.state.axis_locked == 3:
                lock_text = "locked axis: Z"
            state_lines.append(lock_text)

        """ if modal.state.target_face_type == 1:
            sel_vert_text = "ALL"  
        elif modal.state.target_face_type == 3:
            sel_vert_text = "TRIS" 
        elif modal.state.target_face_type == 4:
            sel_vert_text = "QUADS" 
        else:
            sel_vert_text = str(modal.state.target_face_type)+"-GON"
        state_lines.append("Face type: "+sel_vert_text) """

    # if isinstance(state, states.RotateState):
    return state_lines


def set_relative(current, obj, modal):
    target_face = modal.target_object.data.polygons[current.target_face_index]
    ctverts = mesh.get_verts_from_polygon(
        modal.target_object, target_face)

    # visualize_tml4poly(bpy.context, ctverts, current, modal.target_object, 3)

    # ctverts = mesh.sort_tml(ctverts, current.matrix_local,
    #                        modal.target_object.matrix_world)

    ctverts_wc = mesh.converts_verts2wc(modal.target_object, ctverts)
    ctverts_wc = mesh.sort_high_z(ctverts_wc)
    # sverts_wc = mesh.converts_verts2wc(obj, sverts)
    # inside = mesh.geo_inside_poly(tverts_wc, sverts_wc)

    highest_z = mesh.get_highest_z_vert(ctverts_wc)
    draw_lines_3d(bpy.context,
                  [highest_z, (0, 0, 0)], (1, 1, 0), 3)
    edge = mesh.get_highest_z_edge(ctverts_wc)
    draw_lines_3d(bpy.context,
                  [edge[0], edge[1]], (0, 1, 0), 5)

    # mesh.get_midpoint_world(current, ctverts)
    point = mesh.get_midpoint_by_active(current)
    # point = mesh.get_midpoint_world(current, ctverts)

    tri_current = mesh.get_tri_from_point_inside_poly(ctverts_wc, point)
    # print("!!tri :", tri_current)
    if not tri_current:
        return
    # draw_lines_3d(bpy.context,
    #              [tri_current[0], (0, 0, 0)], (0, 1, 1), 5)
    # tri_current_wc = mesh.converts_verts2wc(modal.target_object, tri_current)

    # print("!!!!!!!!!!", tri_current_wc[0])
    draw_poly_3d(bpy.context, tri_current,
                 (1, 0, 1, 0.01), shader=shaders.rgba_3d)
    tri_current_i = mesh.get_indicies(tri_current, ctverts_wc)
    # print("triiiii", tri_current_i)

    target_face_obj = modal.target_object.data.polygons[obj.target_face_index]
    otverts = mesh.get_verts_from_polygon(
        modal.target_object, target_face_obj)
    otverts_wc = mesh.converts_verts2wc(modal.target_object, otverts)
    # otverts = mesh.sort_tml(otverts, obj.matrix_local,
    #                        modal.target_object.matrix_world)
    otverts_wc = mesh.sort_high_z(otverts_wc)
    """ tri_object = mesh.get_verts_by_indicies(tri_current_i, otverts)
    print("!!!!!!!!!!", tri_object)
    tri_object_wc = mesh.converts_verts2wc(modal.target_object, tri_object) """
    tri_object_wc = mesh.get_verts_by_indicies(tri_current_i, otverts_wc)
    draw_poly_3d(bpy.context, tri_object_wc,
                 (0, 1, 1, 0.1), shader=shaders.rgba_3d)

    point2 = mathutils.geometry.intersect_point_tri(
        point, tri_current[0], tri_current[1], tri_current[2])
    # print("p2:", point2)
    if point2 == None:
        return
    draw_lines_3d(bpy.context,
                  [point2, (0, 0, 0)], (1, 0, 1), 5)
    # mesh.get_tri_from_point_inside_poly(ctverts_wc, mesh.get_midpoint_world(current, ctverts)+offset)

    # if obj is current:
    # visualize_poly_indicies(bpy.context, tri_current, 2, (0, -1, 0))
    # visualize_poly_indicies(bpy.context, ctverts_wc, 5)
    # visualize_tml4poly(bpy.context, otverts, obj, modal.target_object, 3)

    """ if obj is current:
        visualize_poly_indicies(bpy.context, mesh.converts_verts2wc(
            modal.target_object, otverts), 2, (0, -1, 0))
    visualize_poly_indicies(bpy.context, ctverts_wc, 5) """

    # visualize_poly_indicies(bpy.context, tri_object_wc, 2, (0, -1, 0))
    # visualize_poly_indicies(bpy.context, tri_current, 5)

    pd = mathutils.geometry.barycentric_transform(
        point2, tri_current[0], tri_current[1], tri_current[2], tri_object_wc[0], tri_object_wc[1], tri_object_wc[2])
    # current_obj = modal.projected_objects[self.object_index]
    draw_lines_3d(bpy.context,
                  [pd, (0, 0, 0)], (1, 0, 1), 5)


def draw_scaling_ui(context, modal):
    """ draw callback funtion from modal grab state """
    inside = True
    tmvp_matrix = context.region_data.perspective_matrix @ modal.target_object.matrix_world

    for obj in modal.projected_objects:

        verts = mesh.get_verts_from_polygon(
            obj, mesh.get_active_polygon(obj))

        # highlight source face of projected objects
        mvp_matrix = context.region_data.perspective_matrix @ obj.matrix_world
        if obj in modal.selected_objects:
            draw_lines_3d(context,
                        mesh.get_line_coords(verts), (1, 0, 1), 1, mvp_matrix)
            draw_poly_3d(context, mesh.get_tri_coords_for_poly( 
                verts), (1, 1, 0, 0.1), mvp_matrix, shader=shaders.rgba_3d)

        # if not modal.state.fit_scale:
            # ... and highlight target faces and check if object is inside
        obj_inside = highlight_target_faces(obj, context, modal, tmvp_matrix)
        inside = obj_inside and inside
        # else : 
        # if modal.state.fit_scale:
        #    obj_inside = True
        #    inside = Truedraw 
        # print("obj",obj.draw 

    # print("all  in:",insdraw 
    # determe the color for target face (red if movement not applyable, green if ok)
    color = (0, 1, 0, 0.1) if inside or not modal.state.limit_to_face else (
        1, 0, 0, 0.1)
    # highlight current target face
    current_obj = modal.projected_objects[modal.object_index]#modal.state.object_index]
    target_face = modal.target_object.data.polygons[current_obj.target_face_index]
    tverts = mesh.get_verts_from_polygon(
        modal.target_object, target_face)
    draw_poly_3d(context, mesh.get_tri_coords_for_poly(
        tverts), color, tmvp_matrix, shader=shaders.rgba_3d)

    # set modal state prop from here for avoiding to calculating double
    modal.state.inside = inside


def draw_grabing_ui(context, modal):
    """ draw callback funtion from modal grab state """
    inside = True
    tmvp_matrix = context.region_data.perspective_matrix @ modal.target_object.matrix_world

    for obj in modal.projected_objects:

        verts = mesh.get_verts_from_polygon(
            obj, mesh.get_active_polygon(obj))

        # highlight source face of projected objects
        mvp_matrix = context.region_data.perspective_matrix @ obj.matrix_world
        if obj in modal.selected_objects:
            draw_lines_3d(context,
                        mesh.get_line_coords(verts), (1, 0, 1), 1, mvp_matrix)    
            draw_poly_3d(context, mesh.get_tri_coords_for_poly(
                verts), (1, 1, 0, 0.1), mvp_matrix, shader=shaders.rgba_3d)

        # ... and highlight target faces and check if object is inside
        obj_inside = highlight_target_faces(obj, context, modal, tmvp_matrix)
        inside = obj_inside and inside

    # determe the color for target face (red if movement not applyable, green if ok)
    color = (0, 1, 0, 0.1) if inside or (
        not modal.state.relative_pos and
        not modal.state.limit_to_face) else (1, 0, 0, 0.1)
    # highlight current target face
    current_obj = modal.projected_objects[modal.object_index]#modal.state.object_index]
    target_face = modal.target_object.data.polygons[current_obj.target_face_index]
    tverts = mesh.get_verts_from_polygon(
        modal.target_object, target_face)
    draw_poly_3d(context, mesh.get_tri_coords_for_poly(
        tverts), color, tmvp_matrix, shader=shaders.rgba_3d)

    # set modal state prop from here for avoiding to calculating double
    modal.state.inside = inside


def highlight_target_faces(obj, context, modal, tmvp_matrix):
    """ highlight all target faces and check if
        current object source face is inside its target face"""
    target_face = modal.target_object.data.polygons[obj.target_face_index]
    tverts = mesh.get_verts_from_polygon(
        modal.target_object, target_face)

    inside = True
    # check if obj is current one
    if obj == modal.projected_objects[modal.object_index]:#state.object_index]:
        # check if all verts of source are inside target face
        sverts = mesh.get_verts_from_polygon(
            obj, mesh.get_active_polygon(obj))
        if type(modal.state) is states.ScaleState and modal.state.fit_scale:
            obj_inside = True
        else:
            obj_inside = mesh.geo_inside_poly(
                tverts, modal.target_object, sverts, obj)
        inside = obj_inside and inside
        # print("obj c",obj.target_face_index,"  in:",inside)
    else:
        # draw all other target faces with way more transparency to only adumbrate them
        color = (0, 1, 0, 0.025)  # inside ok color
        # if absolute pos and limited we need to check if all object are inside their target faces
        if (type(modal.state) is states.ScaleState or not modal.state.relative_pos) and modal.state.limit_to_face:
            sverts = mesh.get_verts_from_polygon(
                obj, mesh.get_active_polygon(obj))
            if type(modal.state) is states.ScaleState and modal.state.fit_scale:
                obj_inside = True
            else:
                obj_inside = mesh.geo_inside_poly(
                    tverts, modal.target_object, sverts, obj)
            inside = obj_inside and inside
            # print("obj",obj.target_face_index,"  in:",inside)
            

            color = (0, 1, 0, 0.025) if obj_inside or (type(modal.state) is not states.ScaleState and
                                                      not modal.state.relative_pos and
                                                      not modal.state.limit_to_face) else (1, 0, 0, 0.1)

        if obj in modal.selected_objects:
            draw_poly_3d(context, mesh.get_tri_coords_for_poly(
                tverts), color, tmvp_matrix, shader=shaders.rgba_3d)

    return inside


def visualize_tml4poly(context, coords, obj, target, thickness, off=(0, 0, 0)):
    """ draw line to tml vert """
    transform = obj.matrix_local
    tml_vert = mesh.get_topmostleft_vert_from_polygon(
        coords, transform, target.matrix_world)
    tml_vert_wc = target.matrix_world @ tml_vert
    draw_lines_3d(context,
                  [tml_vert_wc, off], (1, 1, 1), thickness)


def visualize_poly_indicies(context, coords, thickness, off=(0, 0, 0)):
    """ draw lines in different colors for each vert of the poly to 
        show their order """
    index = 0
    for coord in coords:
        coord = (coord[0]+off[0], coord[1]+off[1], coord[2]+off[2])
        draw_lines_3d(context,
                      [coord, off], color4index(index), thickness)
        index += 1


def color4index(index):
    """ get a different color for each index """
    if index == 0:
        return (1, 0, 0)
    if index == 1:
        return (0, 1, 0)
    if index == 2:
        return (0, 0, 1)
    if index == 3:
        return (1, 1, 0)
    if index == 4:
        return (1, 0, 1)
    if index == 5:
        return (0, 1, 1)
    if index == 6:
        return (0, 0, 0)
    return (1, 1, 1)


def draw_poly_3d(context, coords, color, matrix=None, shader=shaders.rgb_3d):
    """ draw polygons with bgl. Coordinates are given as seperate triangles """
    # if no matrix is given assume coordinates are in world space and use perspective matrix only
    if matrix is None:
        matrix = context.region_data.perspective_matrix

    batch = batch_for_shader(shader, 'TRIS', {"position": coords})

    bgl.glEnable(bgl.GL_BLEND)
    shader.bind()
    shader.uniform_float("MVP_Matrix", matrix
                         )
    shader.uniform_float("color", color)
    batch.draw(shader)

    bgl.glDisable(bgl.GL_BLEND)


def draw_lines_3d(context, coords, color, thickness, matrix=None, shader=shaders.rgb_3d, line_type='LINES'):
    """ draw a lines with bgl. Coordinates are given as pairs of start -and endpoint"""
    # if no matrix is given assume coordinates are in world space and use perspective matrix only
    if matrix is None:
        matrix = context.region_data.perspective_matrix

    batch = batch_for_shader(shader, line_type, {"position": coords})

    bgl.glEnable(bgl.GL_BLEND)

    bgl.glLineWidth(thickness)

    shader.bind()
    shader.uniform_float("MVP_Matrix", matrix)
    shader.uniform_float("color", color)
    batch.draw(shader)

    bgl.glLineWidth(1)  # restore opengl defaults

    bgl.glDisable(bgl.GL_BLEND)

logo_batch = None
logo_inited = False
logo_shader = False
image = None
#blendout = False
LOGO_SPACING = 5

def init_logo(image_name,x, y):
    global logo_batch
    global logo_inited
    global image
    global logo_shader
    logo_inited = True

    #addon_path = bpy.utils.user_resource('SCRIPTS', "addons")
    addon_path = os.path.dirname(os.path.abspath(__file__))
    bpy.data.images.load(addon_path+"/ui/images/"+image_name, check_existing=True)
    image = bpy.data.images[image_name]

    logo_shader = gpu.shader.from_builtin('2D_IMAGE')
    """ logo_batch = batch_for_shader(
        logo_shader, 'TRI_FAN',
        {
            #"pos": ((100, 100), (200, 100), (200, 200), (100, 200)),
            "pos": ((x-offset, y-offset), (x, y-offset), (x, y), (x-offset, y)),
            "texCoord": ((0, 0), (1, 0), (1, 1), (0, 1)),
        },
    ) """

    if image.gl_load():
        raise Exception()


def draw_logo(x, y, size, context):
    if context.window_manager.face2face.blendout: return
    global image
    global logo_batch
    global logo_shader
    #global logo_size
    x -= LOGO_SPACING
    if not logo_inited:
        init_logo("mosaiklogo128x128.png", x, y)
    #size = 128
    offset = size#/2
    logo_batch = batch_for_shader(
        logo_shader, 'TRI_FAN',
        {
            #"pos": ((100, 100), (200, 100), (200, 200), (100, 200)),
            "pos": ((x-offset, y-offset), (x, y-offset), (x, y), (x-offset, y)),
            "texCoord": ((0, 0), (1, 0), (1, 1), (0, 1)),
        },
    )
    bgl.glActiveTexture(bgl.GL_TEXTURE0)
    bgl.glBindTexture(bgl.GL_TEXTURE_2D, image.bindcode)

    logo_shader.bind()
    logo_shader.uniform_int("image", 0)
    #logo_shader.uniform_vec3("pos", (1, 1, 0, 1))
    logo_batch.draw(logo_shader)

def precleanup(context):
    # called at start blendout
    global wireframe_batch
    global wireframe_rotflip_batches
    global rotflip_objects
    global rotflip_needs_to_be_recreated

    wireframe_batch = None
    context.window_manager.f2f_tmp_vars.selection_changed = False
    wireframe_rotflip_batches = []
    rotflip_objects = set()
    rotflip_needs_to_be_recreated = False
    context.window_manager.face2face.do_cleanup = False

def cleanup(context):
    # called in modal cleanup
    global logo_batch
    global logo_inited
    global image
    global logo_shader
    logo_batch = None
    logo_inited = False
    logo_shader = False
    image = None
    context.window_manager.face2face.blendout = False


#bpy.types.SpaceView3D.draw_handler_add(draw, (), 'WINDOW', 'POST_PIXEL')

""" draw.debug_lines = []
draw.debug_lines.append( ( (   org_mid, (0,0,0)  )  , (1,0,0), 5) )
draw.debug_lines.append( ( (   new_mid, (0,0,0)  )  , (0,1,0), 5) )
draw.debug_lines.append( ( (   loc_diff, (0,0,0)  )  , (0,0,1), 5) ) """
