""" This file contains custom mesh operations/transforms for the align faces operator addon"""
import math
from functools import reduce
import mathutils
from mathutils import Vector
import bpy
import bmesh


# Import From Files
if "bpy" in locals():
    import importlib
    from . import helpers
    importlib.reload(helpers)
else:
    from . import helpers


def get_ortho(a_pos, b_pos, c_pos):
    """ calculate a perpendicular vector to a given one """
    if c_pos != 0.0 and -a_pos != b_pos:
        return [-b_pos-c_pos, a_pos, a_pos]
    return [c_pos, c_pos, -a_pos-b_pos]


def clamp(val, minimum, maximum):
    """ limit a value in min and max borders
    TODO: move somewhere else because not realy connect to mesh"""
    if val < minimum:
        return minimum
    if val > maximum:
        return maximum
    return val


def calculate_axis(fnorm, tnorm):
    """ calculate the axis needed to rotate by defined by the normal
    of the plane defined by two vectors """
    axis = fnorm.cross(tnorm)
    # Parallel faces need a new rotation vector
    if axis.length < 1.0e-8:
        axis = Vector(get_ortho(fnorm.x, fnorm.y, fnorm.z))
    return axis


def calculate_angle(fnorm, tnorm):
    """ calculate the angle between to face normals """
    dot = fnorm.normalized().dot(tnorm.normalized())
    dot = clamp(dot, -1.0, 1.0)
    angle = math.acos(dot) + math.pi
    return angle


def adjust_ct_zrot(custom_orient, to_obj, target_face_index):
    """ adjust custom orientation created by blender operator so that the
        orientation is rotated around the z axis to match orientation
        ... for tris  -> by lowest  global z edge
        ... for quads -> by highest global z edge
        ... for ngons -> by highest global z edge """
    target_face_verts = get_verts_from_polygon(
        to_obj, to_obj.data.polygons[target_face_index])
    target_face_verts_wc = converts_verts2wc(to_obj, target_face_verts)

    if len(target_face_verts_wc) == 3:
        edge_points = get_lowest_z_edge(target_face_verts_wc)
    if len(target_face_verts_wc) == 4:
        edge_points = get_highest_z_edge(target_face_verts_wc)
    if len(target_face_verts_wc) > 4:
        edge_points = get_highest_z_edge(target_face_verts_wc)

    # draw.debug_lines.append( ( (   mathutils.Vector(edge_points[0]), mathutils.Vector(edge_points[1]) )  , (1,0,0), 5) )

    traget_axis = mathutils.Vector(
        edge_points[1])-mathutils.Vector(edge_points[0])

    # tpoly = to_obj.data.polygons[target_face_index]
    # tnorm_copy = tpoly.normal.copy()
    # tnorm = to_obj.rotation_quaternion @ tnorm_copy

    """ custom_orient.custom_orientation.matrix = align_orient2axis(
        custom_orient.custom_orientation.matrix, traget_axis, tnorm, edge_points[0]) """
    custom_orient.custom_orientation.matrix = align_orient2axis(
        custom_orient.custom_orientation.matrix, traget_axis)


PARALEL_THRESHHOLD_ANGLE = 0.48


def is_paralel_angle(ang_deg, threshhold=PARALEL_THRESHHOLD_ANGLE):
     """ test if angle between two vecs is nearly paralel """
     return (ang_deg <= 180.0+threshhold and
             ang_deg >= 180.0-threshhold) or \
            (ang_deg <= 0.0+threshhold and
             ang_deg >= 0.0-threshhold)


def align_orient2axisnmew(orient_mat, target_axis, tnorm, t1, is_x_axis=True):
    """ adjust an orientation matrix so either its x or
    its y axis matches an given target axis """
    orient_x = Vector((orient_mat[0][0], orient_mat[1][0], orient_mat[2][0]))
    orient_y = Vector((orient_mat[0][1], orient_mat[1][1], orient_mat[2][1]))
    orient_axis = orient_x if is_x_axis else orient_y

    # draw.debug_lines.append( ( (   mathutils.Vector(t1), +mathutils.Vector(t1)+orient_axis )  , (1,0.4,0), 5) )

    angle = calculate_angle(orient_axis, target_axis)

    axis = tnorm

    rot = mathutils.Matrix.Rotation(angle, 3, axis)
    new_orient = rot @ orient_mat
    return new_orient


def align_orient2axis(orient_mat, target_axis, is_x_axis=True):
    """ adjust an orientation matrix so either its x or
    its y axis matches an given target axis """
    orient_x = Vector((orient_mat[0][0], orient_mat[1][0], orient_mat[2][0]))
    orient_y = Vector((orient_mat[0][1], orient_mat[1][1], orient_mat[2][1]))
    orient_z = Vector((orient_mat[0][2], orient_mat[1][2], orient_mat[2][2]))
    orient_axis = orient_x if is_x_axis else orient_y

    angle = calculate_angle(orient_axis, target_axis)

    """
    if is_paralel_angle(ang_deg):
        orient_axis = orient_y if is_x_axis else orient_x  """

    ang_deg = math.degrees(angle)
    if ang_deg < 200 or ang_deg > 340:
        orient_axis = orient_y if is_x_axis else orient_x

    axis = calculate_axis(orient_axis, target_axis)

    # draw.debug_lines.append( ( (   mathutils.Vector(t1), mathutils.Vector(t1)+axis )  , (0,0.9,0), 5) )
    # draw.debug_lines.append( ( (   mathutils.Vector(t1), mathutils.Vector(t1)+tnorm )  , (0,0,1), 5) )

    rot = mathutils.Matrix.Rotation(angle, 3, axis)
    new_orient = rot @ orient_mat
    return new_orient


def offfset_orient_for_z(orient_mat, angle):
    """ adjust an orientation matrix so either its x ornba
    its y axis matches an given target axis """
    orient_x = Vector((orient_mat[0][0], orient_mat[1][0], orient_mat[2][0]))
    orient_y = Vector((orient_mat[0][1], orient_mat[1][1], orient_mat[2][1]))
    rot = mathutils.Matrix.Rotation(angle, 3, orient_x)
    roty = mathutils.Matrix.Rotation(angle, 3, orient_y)
    new_orient = roty @ rot @ orient_mat
    return new_orient


def rotate_mesh(from_obj, angle, axis):
    """ rotate the geometry of an object by an angle around an given axis """
    mesh = from_obj.data
    b_mesh = bmesh.from_edit_mesh(mesh)

    b_mesh.verts.ensure_lookup_table()
    b_mesh.edges.ensure_lookup_table()
    b_mesh.faces.ensure_lookup_table()

    rot = mathutils.Matrix.Rotation(angle, 3, axis)
    from_obj.inner_mesh_rot = helpers.flatten(rot.copy())

    bmesh.ops.rotate(b_mesh, cent=from_obj.location, matrix=rot,
                     verts=b_mesh.verts, space=from_obj.matrix_world)
    mesh.update()


face2face_orientation_names = []
def delete_custom_transforms(context):
    """ delete all created custom transformations """
    for name in face2face_orientation_names:
        bpy.ops.transform.select_orientation(orientation=name)
        bpy.ops.transform.delete_orientation()

def create_custom_orientation(num):
    """ create a custom orientaion out of the active face and return it"""
    bpy.ops.transform.create_orientation(
        name='Face2Align'+str(num),
        # name="h",
        use=True,
        overwrite=True
    )
    orientation = [
        slot for slot in
        bpy.context.scene.transform_orientation_slots
        if slot.custom_orientation
        and slot.custom_orientation.name == 'Face2Align'+str(num)
    ]
    face2face_orientation_names.append('Face2Align'+str(num))
    return orientation[0]


def set_object2custom_orient(obj, custom_orient):
    """ orients an object to an custom orientation """
    vec_sca = obj.matrix_local.decompose()[2]
    mat_loc = mathutils.Matrix.Translation(obj.matrix_local.to_translation())
    # thanks to machin3 ... mathutils has no correct scaling matrix?
    mat_sca = get_sca_matrix(vec_sca)
    mat_out = mat_loc @ custom_orient.custom_orientation.matrix.to_4x4() @ mat_sca
    obj.matrix_local = mat_out


def get_sca_matrix(scale):
    """
    method from to create scaling matrix manualy
    author: "MACHIN3"
    """
    scale_mx = mathutils.Matrix()
    for i in range(3):
        scale_mx[i][i] = scale[i]
    return scale_mx


def orient2face(from_obj, to_obj, target_face):
    """ orients the the local transformation of the object to be aligned to selected face of
    the target object which is defined by the created custom orientation """
    # maybe bpy.context.object.data.update()
    bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
    bpy.ops.object.mode_set(mode='EDIT', toggle=False)
    deselected_geo = deselect_geometry(from_obj)

    deselected_to_geo = deselect_geometry(to_obj)
    # set active to target face
    old_active_face = get_active_polygon(to_obj)
    
    set_active_polygon_bm(to_obj, target_face)
    # bpy.context.object.data.update()
    # bpy.context.view_layer.update()
    # let blender create a custom orientaion out of the target/active face
    custom_orient = create_custom_orientation(target_face)
    
    # adjust original blender orient so each face follow same convention
    adjust_ct_zrot(custom_orient, to_obj, target_face)

    set_object2custom_orient(from_obj, custom_orient)

    # store 
    # custom_transform = view.get_custom_transform_by_obj(context, obj)
    from_obj.matrix_custom_transform = helpers.flatten(
        custom_orient.custom_orientation.matrix)

    select_geometry(deselected_to_geo, to_obj)
    # set active to the former active face
    set_active_polygon_bm(to_obj, old_active_face.index)

    select_geometry(deselected_geo, from_obj)


def get_active_polygon(obj):
    """ returns the active face from a given object """
    polys = obj.data.polygons
    poly = polys[polys.active]
    return poly

def get_polygon(obj, index):
    """ returns a face from a given object by its index """
    polys = obj.data.polygons
    poly = polys[index]
    return poly

def set_active_polygon_to_first_selected(obj):
    """ returns the active face from a given object """
    mesh = obj.data
    b_mesh = bmesh.from_edit_mesh(mesh)

    face = [f for f in b_mesh.faces if f.select]
    b_mesh.faces.active = face[0]
    face[0].select_set(True)

    if b_mesh.is_wrapped:
        bmesh.update_edit_mesh(mesh, False, False)
    else:
        b_mesh.to_mesh(mesh)
        mesh.update()

def set_active_polygon_bm(obj, poly_index):
    """ returns the active face from a given object """
    mesh = obj.data
    b_mesh = bmesh.from_edit_mesh(mesh)

    face = [f for f in b_mesh.faces if f.index == poly_index]

    b_mesh.faces.active = face[0]
    face[0].select_set(True)

    if b_mesh.is_wrapped:
        bmesh.update_edit_mesh(mesh, False, False)
    else:
        b_mesh.to_mesh(mesh)
        mesh.update()

def get_polygons_by_indicies(indicies, obj):
    """ get a bunch of polys by its indicies """
    return [obj.data.polygons[i] for i in indicies]

def count_selected_polygons(obj):
    return len([p for p in obj.data.polygons if p.select])

def get_selected_polygon(obj):
    """ returns selected face from active object
    asume there is exact 1 """
    return [p for p in obj.data.polygons if p.select][0].index

def select_single_polygon(obj, index):
    """ select the polygon with the given index and deselects all others """
    bpy.ops.object.mode_set(mode='EDIT', toggle=False)
    mesh = obj.data
    b_mesh=bmesh.from_edit_mesh(mesh)
    b_mesh.faces.ensure_lookup_table()

    for face in b_mesh.faces:
        if face.index == index:
            face.select_set(True)
            b_mesh.faces.active = face#face.active_set(True)
        else:
            face.select_set(False)

    if b_mesh.is_wrapped:
        bmesh.update_edit_mesh(mesh, False, False)
    else:
        b_mesh.to_mesh(mesh)
        mesh.update()

    """ for poly in obj.data.polygons:
        if poly.index == index:
            poly.setSelect(True)
        else:
            poly.setSelect(False) """

def get_selected_polygons():
    """ returns all selected faces from active object """

    bpy.ops.object.mode_set(mode='EDIT', toggle=False)

    obj = bpy.context.edit_object
    me = obj.data
    bm = bmesh.from_edit_mesh(me)

    return [f.index for f in bm.faces if f.select]


def rotate_mesh2target_face(from_obj, to_obj, target_face_index):
    # def rotate_mesh2target_face(from_obj, to_obj, fpoly, tpoly):
    """ rotate the mesh from an object so its active face has the
        same orientation than the active face of a target object """
    # i think i may found a bug in blender only for an icosphere at the moment we get a segfault
    #  when accesing a normal of polygon after using the edit object mode toggeling hack
    # so for workaround we store the normal temporary in another variable TODO:find out if ....
    # its normal behaviour (maybe icosphere get reconstructed when changing object mode because
    # it a pure mathematical object internally ... just wildly guessing) otherwise Report bug
    tpoly = to_obj.data.polygons[target_face_index]
    fpoly = get_active_polygon(from_obj)
    tnorm_copy = tpoly.normal.copy()
    fnorm_copy = fpoly.normal.copy()
    # hack to update selection TODO:test if still neeeded and if
    # ...> figure out to do this in a clean way
    # maybe bpy.context.object.data.update()
    bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
    bpy.ops.object.mode_set(mode='EDIT', toggle=False)

    to_obj.rotation_mode = 'QUATERNION'
    tnorm = to_obj.rotation_quaternion @ tnorm_copy
    from_obj.rotation_mode = 'QUATERNION'
    fnorm = from_obj.rotation_quaternion @ fnorm_copy

    axis = calculate_axis(fnorm, tnorm)
    angle = calculate_angle(fnorm, tnorm)
    rotate_mesh(from_obj, angle, axis)

    bpy.context.view_layer.update()
    # hack to update rotation ... TODO figure out to do this in a clean way
    # maybe bpy.context.object.data.update()
    bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
    bpy.ops.object.mode_set(mode='EDIT', toggle=False)


def position_mesh_face2face(from_obj, to_obj, target_face_index):
# def position_mesh_face2face(from_obj, to_obj, fpoly, tpoly):
    """ move an object so his selected face position matches the
    selected face position of another target object  """
    tpoly = to_obj.data.polygons[target_face_index]
    fpoly = get_active_polygon(from_obj)
    fvertices = [from_obj.data.vertices[i].co for i in fpoly.vertices]
    tvertices = [to_obj.data.vertices[i].co for i in tpoly.vertices]

    # calculate midpoints of selected faces in world space
    fbary = get_midpoint_world(from_obj, fvertices)
    tbary = get_midpoint_world(to_obj, tvertices)

    from_obj.location = tbary - (fbary - from_obj.location)

def get_edges_by_vert(verts_of_poly):
    edges = []
    index = 0
    for vert in verts_of_poly:
        index += 1
        if index < len(verts_of_poly):
            edge = (vert, verts_of_poly[index])
            edges.append(edge)
        else:
            edge = (vert, verts_of_poly[0])
            edges.append(edge)
    return edges

def get_edges_relative_to_mid(edges, verts_of_poly, tobj):
    mid = get_midpoint(verts_of_poly)
    rel_edges = []
    for edge in edges:
        rel_edge = (edge[0]-mid, edge[1]-mid)
        rel_edges.append(rel_edge)
    return rel_edges

def get_intersection_point_polyedge_vec(verts_of_poly_wc, vec, vec_start, tobj, obj):
    i_point = None
    edges = get_edges_by_vert(verts_of_poly_wc)
    dist = 9999
    for edge in edges:
        i_points_v_e = mathutils.geometry.intersect_line_line(vec_start, vec, edge[0], edge[1])
        # skip paralell lines with no real intersection point 
        if i_points_v_e is None or (i_points_v_e[0]-i_points_v_e[1]).length_squared > 0.01: 
            continue
        
        dist_v_e = (i_points_v_e[0] - vec_start).length_squared
        if dist_v_e < dist:
            dist = dist_v_e
            i_point = i_points_v_e[0]
    return i_point


def get_midpoint_by_active(obj):
    """ calculate the midpoint for an objects active face """
    verts = get_verts_from_polygon(
        obj, get_active_polygon(obj))
    return get_midpoint_world(obj, verts)


def get_midpoint(vertices):
    """ calculate the midpoint of a set of vertices """
    return reduce(Vector.__add__, vertices) / len(vertices)


def get_midpoint_world(obj, vertices):
    """ calculate the midpoint of a set of vertices
        transform it by the object matrix to world coordinates """
    return obj.matrix_world @ get_midpoint(vertices)


def get_verts_from_polygon(obj, poly):
    """ get vertex local coordinates of a given poly """
    return [obj.data.vertices[i].co for i in poly.vertices]

def get_verts_from_polygon_full(obj, poly):
    """ get verticies of a given poly """
    return [obj.data.vertices[i] for i in poly.vertices]

def get_verts_from_polygon_wc(obj, poly):
    """ get vertex world coordinates of a given poly """
    return [(obj.matrix_world@obj.data.vertices[i].co) for i in poly.vertices]


def faces_have_same_amount_of_verts(poly1_index, poly2_index, obj): 
    """ checks if two faces of the same object are having the same amount of verts """
    poly1_verts = get_verts_from_polygon(
        obj, get_polygon(obj, poly1_index))
    poly2_verts = get_verts_from_polygon(
        obj, get_polygon(obj, poly2_index))
    return len(poly1_verts) == len(poly2_verts) 


def get_tft_type(obj, target_object):
    """ returns the targetface_type of an object: which is essentially
        the amount of verts of its target face on the target object """
    target_face = get_polygon(target_object, obj.target_face_index)
    target_face_verts = get_verts_from_polygon(target_object, target_face)
    return len(target_face_verts)

def get_num_of_verts(obj, face_index):
    poly = get_polygon(obj, face_index)
    verts = get_verts_from_polygon(
        obj, poly)
    return len(verts)

""" def calc_poly_size(obj, poly):
    verts = get_verts_from_polygon(
        obj, poly)
    
     """


def get_topmostleft_vert_index4polygon(from_obj, to_obj):
    """ get vertex index of target face which is the topmostleft one
        if transformed into global xy plane """
    target_face = to_obj.data.polygons[from_obj.target_face_index]
    target_face_verts = get_verts_from_polygon(
        to_obj, target_face)
    transform = from_obj.matrix_local
    vert2wc = to_obj.matrix_world
    tml = get_topmostleft_vert_from_polygon(target_face_verts, transform, vert2wc)
    return target_face_verts.index(tml)


def get_topmostleft_vert_from_polygon(verts, transform, vert2wc):
    """ get vertex which is the topmost (highest y coord in transforms space)
        or if two or more verts have equal high take the leftmost of them
        (highest x coord in transforms space) """
    transform_rot = transform.decompose()[1]
    transform_rot.invert()
    tml_untransed = None
    tml_vert = None
    for vert in verts:
       # print("vert:", vert,"  oy:",vert.x)
        vert_trans = (transform_rot @ vert) + vert2wc.to_translation()
        # draw.draw_lines_3d(bpy.context,
        #          [vert_trans, vert2wc @ vert], (0, 0, 0), 3)
        # if tml_vert is not None :
            # print("vert_trans:", vert_trans, "  y:", (round(tml_vert.y,3) < round(vert_trans.y,3)),"   x:", (round(tml_vert.y,3) == round(vert_trans.y,3) and round(tml_vert.x,3) < round(vert_trans.x,3)))
            # print("vert_trans  !!!!!!!!!!  tmly:", tml_vert.y, "  vty",vert_trans.y,"  oy:",vert.y)
        if tml_vert is None or round(tml_vert.y,3) < round(vert_trans.y,3) or (round(tml_vert.y,3) == round(vert_trans.y,3) and round(tml_vert.x,3) < round(vert_trans.x),3):
            # if tml_vert is not None :print("vert_trans  !!!!!!!!!!  tmly:", tml_vert.y, "  vty",vert_trans.y)
            tml_vert = vert_trans
            tml_untransed = vert
    return tml_untransed


def sort_tml(verts, transform, vert2wc):
    """ sort verts of a polygon in such a way that the topmostleft vert is the first
        in the list and keep the order of verts on the other hand """
    vert_tml = get_topmostleft_vert_from_polygon(verts, transform, vert2wc)
    index = verts.index(vert_tml)
    sorted_verts = verts[index:(len(verts))] + verts[0:index]
    return sorted_verts


def sort_verts_by_index(verts, index):
    """ sort verts of a polygon in such a way that the vert given by index is
         the first in the list and keep the order of verts on the other hand """
    sorted_verts = verts[index:(len(verts))] + verts[0:index]
    return sorted_verts


def sort_high_z(verts_wc):
    """ sort verts of a polygon in such a way that the left vert of the highest z edge
        is the first in the list and keep the order of verts on the other hand """
    highest_z_edge = get_highest_z_edge(verts_wc)
    first = highest_z_edge[0]
    index = verts_wc.index(first)
    sorted_verts = verts_wc[index:(len(verts_wc))] + verts_wc[0:index]
    return sorted_verts


def get_opposite_edge(vert, tri_verts):
    """ gets the opposite edge of a given point of an triangle """
    index = tri_verts.index(vert)
    if index == 0:
        return (tri_verts[2], tri_verts[1])
    if index == 1:
        return (tri_verts[0], tri_verts[2])
    if index == 2:
        return (tri_verts[1], tri_verts[0])


def get_lowest_z_edge(tri_verts):
    """ gets the opposite edge of a given point of an triangle """
    highest_z = get_highest_z_vert(tri_verts)
    return get_opposite_edge(highest_z, tri_verts)


def get_highest_z_edge(poly_verts):
    """ gets the the two vectors of the edge of from a polygon which contains the highest z vert in world space
        and its orientation more far away from the z axis than the other edge containing highest z vert 
        poly_verts: just the coordinates in world space"""
    highest_z = get_highest_z_vert(poly_verts)
    edge_cw = get_clockwise_edge(highest_z, poly_verts)
    edge_ccw = get_counterclockwise_edge(highest_z, poly_verts)
    """ draw.draw_lines_3d(bpy.context,
                       [edge_cw[0], edge_cw[1]], (1, 1, 0), 3)
    draw.draw_lines_3d(bpy.context,draw.
                       [edge_ccw[0], edge_ccw[1]], (1, 1, 0), 3) """

    z_axis = mathutils.Vector((0, 0, 1))
    acw = calculate_angle(mathutils.Vector(edge_cw[0])-mathutils.Vector(edge_cw[1]), z_axis)
    acw2 = calculate_angle(mathutils.Vector(edge_cw[1])-mathutils.Vector(edge_cw[0]), z_axis)
    accw = calculate_angle(mathutils.Vector(edge_ccw[0])-mathutils.Vector(edge_ccw[1]), z_axis)
    accw2 = calculate_angle(mathutils.Vector(edge_ccw[1])-mathutils.Vector(edge_ccw[0]), z_axis)

    acw = math.degrees(acw)
    accw = math.degrees(accw)
    acw2 = math.degrees(acw2)
    accw2 = math.degrees(accw2)

    acw = min(acw, acw2)
    accw = min(accw, accw2)

    if  acw > accw:
        return edge_cw
    return edge_ccw


def get_highest_z_edge_indicies(obj, face_index):
    """ gets the edge of from a polygon which contains the highest z vert in world space and its
        orientation more far away from the z axis than the other edge containing highest z vert
        and return edge in form of vert indicies """
    # determe world coords of face
    face = obj.data.polygons[face_index]
    face_verts_f = get_verts_from_polygon_full(
        obj, face )
    face_verts_c = get_verts_from_polygon(
        obj, face )
    poly_verts = converts_verts2wc(obj, face_verts_c)  

    # get the highest z edge (as two vectors in world coords)
    z_edge = get_highest_z_edge(poly_verts)

    # determ the indicies of the verts from edge and return them
    return (face.vertices[poly_verts.index(z_edge[0])],face.vertices[poly_verts.index(z_edge[1])])


def get_clockwise_edge(vert, poly_verts):
    """ gets the edge in clockwise direction of a given point of an polygon """
    index = poly_verts.index(vert)
    cw_index = index - 1
    if cw_index < 0:
        cw_index = len(poly_verts) - 1
    return (poly_verts[cw_index], vert)


def get_counterclockwise_edge(vert, poly_verts):
    """ get edge in counterclockwise direction of a given point of an polygon """
    index = poly_verts.index(vert)
    ccw_index = index + 1
    if ccw_index >= len(poly_verts):
        ccw_index = 0
    return (vert, poly_verts[ccw_index])   


def get_highest_z_vert(verts):
    """ the highest z vert in world space from a bunch of verts """
    highest_z_vert = None
    for vert in verts:
        if highest_z_vert is None or round(highest_z_vert.z, 3) < round(vert.z, 3) or (round(highest_z_vert.z, 3) == round(vert.z, 3) and round(highest_z_vert.y, 3) < round(vert.y, 3)) or (round(highest_z_vert.z, 3) == round(vert.z, 3) and round(highest_z_vert.y, 3) == round(vert.y, 3) and round(highest_z_vert.x, 3) < round(vert.x, 3)):
            highest_z_vert = vert
    return highest_z_vert


def get_indicies(verts, poly):
    """ get the indicies of a bunch of verts in a polygon """
    indicies = []
    for vert in verts:

        index = 0
        for vertp in poly:
            if vert is vertp:
                indicies.append(index)
            index += 1
    return indicies


def get_verts_by_indicies(indicies, vertices):
    """ get a bunch of verts by its indicies """
    return [vertices[i] for i in indicies]


def converts_verts2wc(obj, verts):
    """ convert a set of local vertex coordinates
        by the object matrix to world coordinates """
    verts_wc = []
    for vert in verts:
        # vert_wc = obj.matrix_world @ vert
        # verts_wc.append(vert_wc)
        vert_wc = convert_vert2wc(obj, vert)
        verts_wc.append(vert_wc)
    return verts_wc

def convert_vert2wc(obj, vert):
    """ convert a set of local vertex coordinates
        by the object matrix to world coordinates """
    vert_wc = obj.matrix_world @ vert
    return vert_wc

def converts_verts_wc2local(obj, verts):
    """ convert a set of local vertex coordinates
        by the object matrix to world coordinates """
    verts_local = []
    for vert in verts:       
        vert_local = convert_vert_wc2local(obj, vert)
        verts_local.append(vert_local)
    return verts_local

def convert_vert_wc2local(obj, vert):
    """ convert a set of local vertex coordinates
        by the object matrix to world coordinates """
    wc_matrix_inv = obj.matrix_world.copy()
    wc_matrix_inv.invert()
    vert_local = wc_matrix_inv @ vert
    return vert_local


def get_faces_from_loop(loop):
    """ get the face indicies of faces starting from a loop/edge in 1 direction
        and going over its oppenent edge to the adjacent face and so on until
        it reaches the border of the mesh or the starting face again """
    faces = set()
    start_face = loop.face
    starting = True
    manifold = True
    border_reached = False

    while (manifold and start_face is not loop.face and len(loop.face.verts)==4 and not border_reached) or starting:

        starting = False
        # loop.face.select_set(True)
        faces.add(loop.face.index)

        # get opposite edge/loop
        loop = loop.link_loop_next.link_loop_next
        
        # get loops/loop(i manifold) from adjacent face
        opp_rad_next = loop.link_loop_radial_next
        opp_rad_prev = loop.link_loop_radial_prev

        # the edge opposite edge/loop is manifold when only 1 adjacent loop/face exists
        manifold = (opp_rad_next is opp_rad_prev )
        # check if border of mesh reached ...-> when no adjacent loop/face exists
        border_reached = manifold and (opp_rad_next is loop)

        # set current loop to opposite adjacent loop
        loop = opp_rad_next
    
    return faces



def get_loop_index_for_edge(bmesh_face, z_edge):
    """ determe the loop index by an edge (given through its verts indicies) """
    loop_index = 0
    for loop in bmesh_face.loops:
        if loop.vert.index in z_edge and loop.edge.other_vert(loop.vert).index in z_edge:
            return loop_index
        loop_index += 1
    return -1



def get_face_loop_faces(obj, index, direction):
    """ get the face indicies of face loop starting from a face in a given
        direction(1 or 0 meaning something like (left/x and right/x) or (up/y and down/y)
        walking in both sides of the direction in case of an unfinished loop """
    # get bmesh data and ensure it
    mesh = obj.data
    b_mesh = bmesh.from_edit_mesh(mesh)

    b_mesh.verts.ensure_lookup_table()
    b_mesh.edges.ensure_lookup_table()
    b_mesh.faces.ensure_lookup_table()

    # create set to store indicies of faces from the loop
    loop_faces = set()

    # the target face of the selected projected object acts as starting point
    start_face = b_mesh.faces[index]

    # determ correct diretion via highest z_edge
    z_edge = get_highest_z_edge_indicies(obj, index)
    switch_direction = get_loop_index_for_edge(start_face, z_edge)%2 == 0
    if switch_direction:
        direction = 0 if direction == 1 else 1

    # draw.debug_lines.append( ( (   obj.location+start_face.verts[0].co, (0,0,0)  )  , (1,0,0), 5) )
    # determ faces from loop by starting to both sides of a given direction 
    if len(start_face.verts) == 4: # only quads have loops
        loop_faces.update(get_faces_from_loop(start_face.loops[3-direction]))
        loop_faces.update(get_faces_from_loop(start_face.loops[1-direction]))

    if b_mesh.is_wrapped:
        bmesh.update_edit_mesh(mesh, False, False)
    else:
        b_mesh.to_mesh(mesh)
        mesh.update()

    return loop_faces

def deselect_geometry(obj):
    """ deselect all verts, edges and faces of an object and return their indicies
    as an triple(3-tupel) of Lists of their indicies"""
    mesh = obj.data
    b_mesh = bmesh.from_edit_mesh(mesh)

    selected_verts = [v for v in b_mesh.verts if v.select]
    selected_edges = [e for e in b_mesh.edges if e.select]
    selected_faces = [f for f in b_mesh.faces if f.select]

    for vert in selected_verts:
        vert.select_set(False)
    for edge in selected_edges:
        edge.select_set(False)
    for face in selected_faces:
        face.select_set(False)

    if b_mesh.is_wrapped:
        bmesh.update_edit_mesh(mesh, False, False)
    else:
        b_mesh.to_mesh(mesh)
        mesh.update()

    selected_verts = [v.index for v in selected_verts]
    selected_edges = [e.index for e in selected_edges]
    selected_faces = [f.index for f in selected_faces]
    return (selected_verts, selected_edges, selected_faces)


def select_geometry(data, obj):
    """ select verts, edges and faces of an object defined by their indicies inside
    data (a 3-tupel of respective lists) """
    selected_verts = data[0]
    selected_edges = data[1]
    selected_faces = data[2]

    bpy.ops.object.mode_set(mode='EDIT', toggle=False)
    mesh = obj.data
    b_mesh = bmesh.from_edit_mesh(mesh)

    for vert in b_mesh.verts:
        if vert.index in selected_verts:
            vert.select_set(True)
    for edge in b_mesh.edges:
        if edge.index in selected_edges:
            edge.select_set(True)
    for face in b_mesh.faces:
        if face.index in selected_faces:
            face.select_set(True)

    if b_mesh.is_wrapped:
        bmesh.update_edit_mesh(mesh, False, False)
    else:
        b_mesh.to_mesh(mesh)
        mesh.update()


def geo_inside_poly(verts_of_poly, poly_obj, verts, obj):
    """ determe if the projection of a bunch of vertices onto the plane
        of the tris of a target face is inside that target face.
        Can used for example to check if a full poly projection is inside another poly
        all verts are going to be converted from its local to same world space """
    tverts_wc = converts_verts2wc(poly_obj, verts_of_poly)
    sverts_wc = converts_verts2wc(obj, verts)

    # printprint("tverts_wc",tverts_wc)
    # print("sverts_wc",sverts_wc)

    return geo_inside_poly_wc(tverts_wc, sverts_wc)

def geo_inside_poly_wc(coords_of_poly, vertices):
    """ determe if the projection of a bunch of vertices onto the plane
        of the tris of a target face is inside that target face.
        Can used for example to check if a full poly projection is inside another poly
        all verts need to be in the same world space """
    for vert in vertices:
        if not point_inside_poly(coords_of_poly, vert):
            # print("outside vert",vert)
            return False
    return True


def get_tri_from_point_inside_poly(coords_of_poly, point):
    """ determe if the projection of a point onto the plane of the tris of a face
        is inside the polygon """
    if len(coords_of_poly) == 3:
        return coords_of_poly if inside_tri(coords_of_poly, point) else False
    if len(coords_of_poly) == 4:
        return get_inside_quad(coords_of_poly, point)
    if len(coords_of_poly) > 4:
        return get_inside_ngon(coords_of_poly, point)


def get_inside_quad(coords_of_quads, point):
    """ determe if the projection of a point onto the plane of the tris of a quad
        is inside the polygon """
    tris = get_tri_arr_coords_for_quad(coords_of_quads)
    for tri in tris:
        if inside_tri(tri, point):
            return tri
    return False


def get_inside_ngon(coords_of_ngon, point):
    """" determe if the projection of a point onto the plane of the tris of an ngon
        is inside the polygon """
    tris = get_tri_arr_coords_for_ngon(coords_of_ngon)
    for tri in tris:
        if inside_tri(tri, point):
            return tri
    return False


def point_inside_poly(coords_of_poly, point):
    """ determe if the projection of a point onto the plane of the tris of a face
        is inside the polygon """
    if len(coords_of_poly) == 3:
        return inside_tri(coords_of_poly, point)
    if len(coords_of_poly) == 4:
        return inside_quad(coords_of_poly, point)
    if len(coords_of_poly) > 4:
        return inside_ngon(coords_of_poly, point)


def inside_ngon(coords_of_ngon, point):
    """" determe if the projection of a point onto the plane of the tris of an ngon
        is inside the polygon """
    tris = get_tri_arr_coords_for_ngon(coords_of_ngon)
    for tri in tris:
        if inside_tri(tri, point):
            return True
    return False


def inside_quad(coords_of_quads, point):
    """ determe if the projection of a point onto the plane of the tris of a quad
        is inside the polygon """
    tris = get_tri_arr_coords_for_quad(coords_of_quads)
    # print("tries",tris)
    for tri in tris:
        if inside_tri(tri, point):
            return True
    return False


def inside_tri(coords_of_tri, point):
    """ determe if the projection of a point onto the plane of the tris is inside it """
    return mathutils.geometry.intersect_point_tri(point,
                                                  coords_of_tri[0],
                                                  coords_of_tri[1],
                                                  coords_of_tri[2])


def get_tri_arr_coords_for_quad(vertices):
    """ convert a quad into tri coords, put each tri in a seperate array, for using
        the tris virtually only without creating them by triangulate for example """
    coords = [[vertices[0], vertices[1], vertices[2]],
              [vertices[2], vertices[3], vertices[0]]]
    return coords


def get_tri_arr_coords_for_ngon(vertices):
    """ convert a quad into tri coords, put each tri in a seperate array, for using
        the tris virtually only without creating them by triangulate for example """
    vertices = get_tri_coords_for_ngon(vertices)
    coords = []
    for i in range(0, int(len(vertices)/3)):
        tri = []
        tri.append(vertices[0+(3*i)])
        tri.append(vertices[1+(3*i)])
        tri.append(vertices[2+(3*i)])
        coords.append(tri)
    return coords


def get_tri_coords_for_poly(vertices):
    """ converts an convex ngon into tri coords convention notation
        virtually triangulate the poly and store verts in an array
        where each tripple forms a triangle: used for example for drawing
        remark:drawing could also be done with passing indicies to batch_for_shader"""
    if len(vertices) == 3:
        return vertices
    if len(vertices) == 4:
        return get_tri_coords_for_quad(vertices)
    if len(vertices) > 4:
        return get_tri_coords_for_ngon(vertices)


def get_tri_coords_for_ngon(vertices):
    """ convert an convex ngon into tri coords convention via fan-triangulate """
    coords = [vertices[0], vertices[1], vertices[2]]
    index = 0
    indicies = [0, 1, 2]
    for vert in vertices:
        # if index != 0 and index != 1 and index != len(vertices)-1:
        if index not in (0, 1, len(vertices)-1):

            coords.append(vertices[0])
            coords.append(vert)
            coords.append(vertices[index+1])
            indicies.append(0)
            indicies.append(index)
            indicies.append(index+1)
        index += 1
    return coords


def get_tri_coords_for_quad(vertices):
    """ convert a quad into tri coords convention for drawing """
    coords = [vertices[0], vertices[1], vertices[2],
              vertices[2], vertices[3], vertices[0]]
    return coords


def get_line_coords(vertices):
    """ convert a polygon vertices into line seperate coords convention """
    coords = []
    index = 0
    for vert in vertices:
        coords.append(vert)
        if index != 0:
            coords.append(vert)
        index += 1
    coords.append(vertices[0])
    return coords
