""" Base class for an modal state """
import bpy
import math
import mathutils

if "bpy" in locals():
    import importlib
    from .. import mesh
    importlib.reload(mesh)
    from .. import view
    importlib.reload(view)
    from .. import helpers
    importlib.reload(helpers)
    from .. config import keymap
    importlib.reload(keymap)
else:
    # pass
    from .. import mesh
    from .. import view
    importlib.reload(view)
    from .. import helpers
    from .. config import keymap


class ModalState:
    """ Base class for an modal state """
    name = "Base"

    target_face_types = set()
    target_face_type = 1  # 1 = all types, 3...n = tri....ngon

    view_button_pressed = False

    select_mode = 0  # 1 manually, 2 object type, 3, circle, 4 box, 6 faceloop

    mouse_start = (0, 0)

    former_status_text = None

    # properties for setting values by num keys
    num_enter_mode = False
    num_value = 0
    decimal_place = 0
    period_used = False

    def map_string2num(self, num_in_word):
        """ align view to the next object  """
        num = -1
        if num_in_word == "ONE":
            num = 1
        if num_in_word == "TWO":
            num = 2
        if num_in_word == "THREE":
            num = 3
        if num_in_word == "FOUR":
            num = 4
        if num_in_word == "FIVE":
            num = 5
        if num_in_word == "SIX":
            num = 6
        if num_in_word == "SEVEN":
            num = 7
        if num_in_word == "EIGHT":
            num = 8
        if num_in_word == "NINE":
            num = 9
        if num_in_word == "ZERO":
            num = 0
        return num

    def select_by_ngon(self, face_type, additive, modal, context, need_2b_selected=False):
        """ select objects of a specific target face type (tri,quad...) """
        if face_type == 1:  # select all
            self.select_all(modal, context)
            return
        # no such type ... so do nothing
        if face_type not in modal.target_face_types:
            return

        if need_2b_selected:  # select objects only if they were already selected before
            new_selected = [
                obj for obj in modal.projected_ngons[face_type] if obj in modal.selected_objects]

        # if not additive and in the special case of grabbing where in relative positioning mode all
        # former selected obj are going to be removed
        # (isinstance(self, GrabState) and self.relative_pos):
        if not additive:
             modal.selected_objects = set()
        if not need_2b_selected:
             new_selected = modal.projected_ngons[face_type]

        # update set and select the determed ones
        modal.selected_objects.update(new_selected)
        self.select_selected(modal, context)


    def select_by_ngonole(self, face_type, additive, modal, context, need_2b_selected=False):
        """ select objects of a specific target face type (tri,quad...) """
        if face_type == 1:  # select all
            self.select_all(modal, context)
            return
        # no such type ... so do nothing
        if face_type not in modal.target_face_types:
            return

        if need_2b_selected:  # select objects only if they were already selected before
            new_selected = [
                obj for obj in modal.projected_ngons[face_type] if obj in modal.selected_objects]

        # if not additive and in the special case of grabbing where in relative positioning mode all
        # former selected obj are going to be removed
        # (isinstance(self, GrabState) and self.relative_pos):
        if not additive:
             modal.selected_objects = set()
        if not need_2b_selected:
             new_selected = modal.projected_ngons[face_type]

        # update set and select the determed ones
        modal.selected_objects.update(new_selected)
        self.select_selected(modal, context)


    def select_by_loop(self, context, modal, loop_faces, additive):
        """ select all objects in a set of target faces """
        # collect loop objects by their target face indicies
        objects_on_loop = modal.get_projected_objects_by_tft(loop_faces)

        if not additive:
             modal.selected_objects = set()

        # target faces on a loop are all quads, so no special relative positioning case
        modal.selected_objects.update(objects_on_loop)

        # select the determed ones
        self.select_selected(modal, context)

    def select_by_type(self, object_type, additive, modal, context):
        """ select all projected objects of the given object type """
        # remove all former selected if the new ones should not be added to them
        if not additive:
             modal.selected_objects = set()
        # handle special case of grabbing where in relative positioning mode all
        # target faces has to have the same amount of verts to calc barycentric co.
        if (self.name == "MOVE" and self.relative_pos):
            type_and_same_ngon = []
            for obj in modal.projected_types[object_type]:
                target_face_type = mesh.get_tft_type(obj, modal.target_object)
                if self.target_face_type == target_face_type:
                    type_and_same_ngon.append(obj)
            modal.selected_objects.update(type_and_same_ngon)
        else:
            modal.selected_objects.update(modal.projected_types[object_type])
         # select the determed ones
        self.select_selected(modal, context)

    def select_all(self, modal, context):
        """ select all projected objects """
        modal.selected_objects.update(modal.projected_objects)
        self.select_selected(modal, context)


    def deselect_by_type(self, object_type, modal, context):
        """ deselect all objects of the given object type """    
        self.deselect_by_objects(context, modal, modal.projected_types[object_type])

    def deselect_by_ngon(self, face_type, modal, context):
        """ deselect all objects which target face is of the given type(tri,quad...)  """
        # deselect all (literally .... one keeps always selected)
        if face_type == 1:
            modal.selected_objects = set()
            self.select_selected(modal, context)
            return
        # no such type ... so do nothing
        if face_type not in modal.target_face_types:
            return
        # remove the specific ngon set
        self.deselect_by_objects(context, modal, modal.projected_ngons[face_type])
    
    def deselect_by_target_faces(self, context, modal, loop_faces):
        """ deselect all objects in a set of target faces indicies """   
        objects_on_loop = []
        for index in loop_faces:
            for obj in modal.projected_objects:#context.selected_objects:
                if obj.target_face_index == index:
                    # obj.select_set(True)
                    objects_on_loop.append(obj)

        self.deselect_by_objects(context, modal, objects_on_loop)

    def deselect_by_objects(self, context, modal, objects):
        """ deselect all objects in a set of objects """
        rest = [obj for obj in modal.selected_objects if obj not in objects]
        modal.selected_objects = set()
        modal.selected_objects.update(rest)
        self.select_selected(modal, context)

    def select_selected(self, modal, context):
        """ select all objects also in blender which have been determed to
            be selected  in the set modal.selected_objects """
        # deselect all projected objects ....
        for obj in modal.projected_objects:
            obj.select_set(False)
        # select only selected projected objects
        for obj in modal.selected_objects:
            obj.select_set(True)

        # handle 0 selected... 0 selected in this context makes no sense so select something
        if len(modal.selected_objects) == 0:
            first_obj = modal.projected_objects[0]
            first_obj.select_set(True)
            modal.selected_objects.add(first_obj)
            # adjust index and target_face_type
            target_face_type = mesh.get_tft_type(first_obj, modal.target_object)
            self.target_face_type = target_face_type
            modal.object_index = 0#modal.projected_objects.index()

        # if current object not inside selected set to first
        if not modal.projected_objects[modal.object_index] in modal.selected_objects:
            # get first object of set without removing it
            first_selected_obj = next(iter(modal.selected_objects)) 
            modal.object_index = modal.projected_objects.index(first_selected_obj)

        context.view_layer.objects.active = modal.projected_objects[modal.object_index]
        if context.window_manager.face2face.auto_view:
            self.align_view2obj(context, modal)

        #draw.selection_changed = True
        #draw.update_rotflip_objects(modal.selected_objects)
        context.window_manager.f2f_tmp_vars.selection_changed = True
        context.window_manager.f2f_tmp_vars.update_rotflip_objects = True
        context.window_manager.f2f_tmp_vars.rotflip_objects = []
        context.window_manager.f2f_tmp_vars.rotflip_objects += modal.selected_objects


    def set_target_face_type(self, context, modal, face_type_string):
        """ set the type of target face (tries, quads, etc) which were selected last?
            and automatically align view if wanted  """
        # determe next object index
        num = self.map_string2num(face_type_string)
        if num in modal.target_face_types:
            self.target_face_type = num
            self.reset(context, modal)
        if num == 1:
            self.target_face_type = num
            self.reset(context, modal)


    def obj_has_current_tft(self, obj, modal):
        """ check if object has currently selected target face type """
        tpoly = mesh.get_polygon(modal.target_object, obj.target_face_index)
        tpoly_verts = mesh.get_verts_from_polygon(
            modal.target_object, tpoly)
        # same or 1 for all types
        return self.target_face_type in (len(tpoly_verts), 1)


    def view_next(self, context, modal):
        """ align view to the previous object  """
        self.next_object_index(modal)
        current_obj = modal.projected_objects[modal.object_index]
        if current_obj in modal.selected_objects:
            if context.window_manager.face2face.auto_view:
                self.align_view2obj(context, modal)
            context.view_layer.objects.active = current_obj
        else:
            # try next projected object
            self.view_next(context, modal)

    def next_object_index(self, modal):
        """ determe and set next object index  """
        modal.object_index += 1
        if modal.object_index >= len(modal.projected_objects):
            modal.object_index = 0

    def view_prev(self, context, modal):
        """ align view to the previous object  """
        self.prev_object_index(modal)
        current_obj = modal.projected_objects[modal.object_index]
        if current_obj in modal.selected_objects:
            if context.window_manager.face2face.auto_view:
                self.align_view2obj(context, modal)
            context.view_layer.objects.active = current_obj
        else:
            # try next projected object
            self.view_prev(context, modal)

    def prev_object_index(self, modal):
        """ determe and set previous object index  """
        modal.object_index -= 1
        if modal.object_index < 0:
            modal.object_index = len(modal.projected_objects) - 1

    def align_view2obj_z(self, context, modal):
        """ align view to the actual object  """
        obj = modal.projected_objects[modal.object_index]
        view.align_view2custom_transform_target_face_z(
            context, obj, modal.target_object)

    def align_view2obj(self, context, modal):
        """ align view to the actual object  """
        obj = modal.projected_objects[modal.object_index]
        view.align_view2custom_transform_target_face(
            context, obj, modal.target_object)

    def reset(self, context, modal):
        """ reset everything to its original state at the beginning of state """
        # for obj in context.selected_objects:
        #    obj.matrix_local = mathutils.Matrix(obj.matrix_local_org)

    def reset2start(self, modal):
        """ reset everything to its original state at the beginning of state """
        for obj in modal.projected_objects:
            obj.matrix_local = mathutils.Matrix(obj.matrix_local_startorg)
            # update the stored scaling per object
            obj.scaling = obj.matrix_local_startorg.decompose()[2]
            # reset the stored angle
            obj.rotation = 0 
        
        # TODO may allow to keep those varying settings
        modal.rotation_variation_strength = 0
        modal.scale_variation_strength = 0
        modal.scale_relative_strength = 0

    def count_sel(self):
        return len(bpy.context.selected_objects) 
  

    def update_selection(self, context, modal):
        """ this functions ensure the correct things were selected in the 
            different selection mehtods, because we use using built in selection ops
            could maybe select objects outside the projected ones.
            And sets the modal lists accordingly """
        # check and reset if object outside projected ones was selected
        for obj in context.selected_objects:
            if not obj in modal.projected_objects:
                obj.select_set(False)

        # some thing outside projected objects was selected so set active back to former
        if not context.active_object in modal.projected_objects:
            context.view_layer.objects.active = modal.projected_objects[modal.object_index]

        # only updated if something relevant was selected
        if len(context.selected_objects) > 0:
            # update selection and active object
            modal.selected_objects = set()
            modal.selected_objects.update(context.selected_objects.copy())
            modal.object_index = modal.projected_objects.index(context.active_object)      
        
        self.select_selected(modal, context)
        return

    def determ_type(self, context, modal):

        # check and reset if object outside projected was selected
        for obj in context.selected_objects:
            if not obj in modal.projected_objects:
                obj.select_set(False)
                self.select_selected(modal, context)
                # return

        # update selection and active object
        obj_type = context.active_object.object_type
        context.view_layer.objects.active = modal.projected_objects[modal.object_index]
        return obj_type
        
    # select_types = set()
    def handle_events_pre(self, context, event, modal):
        """ handle events bevor handling the specific state """
        # target face type select 
        if event.type in {"ONE", "THREE", "FOUR", "FIVE", "SIX", "SEVEN", "EIGHT", "NINE", "ZERO"}:
            if event.value == 'RELEASE':
                
                self.set_target_face_type(context, modal, event.type)
                if event.ctrl:
                    self.deselect_by_ngon(self.map_string2num(event.type), modal, context)
                else:
                    self.select_by_ngon(self.map_string2num(event.type), event.shift, modal, context)
                
                modal.report({'INFO'}, event.type)
            return {'RUNNING_MODAL'}
   

        if self.select_mode == 2: # select object type
            if context.window_manager.face2face.is_select(event.type):#event.type in context.window_manager.face2face.right_click_select:
                if event.value == 'RELEASE':
                    self.select_mode = 0
                    # bpy.ops.view3d.select(deselect_all=True, location=(event.mouse_region_x, event.mouse_region_y))#event.shift [modal.object_index]
                    result = bpy.ops.view3d.select(extend=True, location=(event.mouse_region_x, event.mouse_region_y))
                    context.view_layer.update()
                    result = self.catch_wrong_selection(result, modal, context)
                    if not result is None:
                        return result

                    # print("context.active_object  oooooo:::",context.active_object," selected:",context.active_object.select_get())
                    # click_in_projected_objects = context.view_layer.objects.active in modal.projected_objects and context.view_layer.objects.active.select_get()
                    # print("click_in_projected_objects::::",click_in_projected_objects)    
                          
                    # self.update_selection(context, modal) 
                    # if click_in_projected_objects:
                    obj_type = self.determ_type(context, modal)                    
                    if event.ctrl:
                        self.deselect_by_type(obj_type, modal, context)
                    else:
                        self.select_by_type(obj_type, event.shift, modal, context)
                    modal.report({'INFO'}, "object type selectected")
                    # else:
                    #    modal.report({'INFO'}, "object type selectection cancelled")
                    self.select_selected(modal, context) 
                    # modal.report({'INFO'}, "object type selected:"+str(obj_type))
                    self.mouse_start = (event.mouse_region_x,
                             event.mouse_region_y)
                    context.area.tag_redraw()
                    context.window.cursor_set("DEFAULT")
                    if not self.former_status_text is None:
                        modal.set_status_text(self.former_status_text, False)

            if event.type in {'ESC'} or context.window_manager.face2face.is_non_select(event.type):
                if event.value == 'RELEASE':
                    self.select_mode = 0
                    context.window.cursor_set("DEFAULT")
                    modal.report({'INFO'}, "object type selection cancelled")
                    context.area.tag_redraw()
                    if not self.former_status_text is None:
                        modal.set_status_text(self.former_status_text, False)
                    self.mouse_start = (event.mouse_region_x,
                                event.mouse_region_y)
            return {'RUNNING_MODAL'}

        if self.select_mode == 3: #circle select
            if event.type == 'LEFTMOUSE':
                if event.value == 'RELEASE':
                    # self.select_mode = 0
                    context.window.cursor_set("DEFAULT")
                    if not event.shift:
                         bpy.ops.view3d.select_circle(x=event.mouse_region_x, y=event.mouse_region_y, radius=context.window_manager.f2f_tmp_vars.circle_selecting[2], wait_for_input=False, mode='ADD')
                        
                    else:
                        bpy.ops.view3d.select_circle(x=event.mouse_region_x, y=event.mouse_region_y, radius=context.window_manager.f2f_tmp_vars.circle_selecting[2], wait_for_input=False, mode='SUB')
                    self.update_selection(context, modal)
                    # modal.report({'INFO'}, "circle selection done")
                    context.area.tag_redraw()

            if event.type in {'ESC'} or event.type == 'RIGHTMOUSE':
                if event.value == 'RELEASE':
                    context.window_manager.f2f_tmp_vars.circle_selecting = []
                    self.select_mode = 0
                    if not self.former_status_text is None:
                        modal.set_status_text(self.former_status_text, False)
                    modal.report({'INFO'}, "circle selection left")
                    context.area.tag_redraw()
                    

            if event.type in {'WHEELUPMOUSE'}:
                context.window_manager.f2f_tmp_vars.circle_selecting[2] = min(250, context.window_manager.f2f_tmp_vars.circle_selecting[2] + 5) 
                context.area.tag_redraw()

            if event.type in {'WHEELDOWNMOUSE'}:
                context.window_manager.f2f_tmp_vars.circle_selecting[2] =  max(25 ,context.window_manager.f2f_tmp_vars.circle_selecting[2] - 5)
                context.area.tag_redraw()

            if event.type == 'MOUSEMOVE':
                context.area.tag_redraw()
                context.window_manager.f2f_tmp_vars.circle_selecting = [event.mouse_region_x, event.mouse_region_y, context.window_manager.f2f_tmp_vars.circle_selecting[2]]
            self.mouse_start = (event.mouse_region_x,
                             event.mouse_region_y)
            return {'RUNNING_MODAL'}

        if self.select_mode == 4: #box select
            if event.type == 'LEFTMOUSE':
                if event.value == 'RELEASE':
                    # context.window.cursor_set("DEFAULT")
                    xmin = min(context.window_manager.f2f_tmp_vars.box_selecting[0],event.mouse_region_x)
                    xmax = max(context.window_manager.f2f_tmp_vars.box_selecting[0],event.mouse_region_x)
                    ymin = min(context.window_manager.f2f_tmp_vars.box_selecting[1],event.mouse_region_y)
                    ymax = max(context.window_manager.f2f_tmp_vars.box_selecting[1],event.mouse_region_y)
                    if not event.shift:
                        bpy.ops.view3d.select_box(xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax, 
                                                  wait_for_input=False, mode='ADD')
                        modal.report({'INFO'}, "BOX SELECTION ADDITIVE:")
                    else:
                        bpy.ops.view3d.select_box(xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax, 
                                                  wait_for_input=False, mode='SUB')
                        """ bpy.ops.view3d.select_box(xmin=context.window_manager.f2f_tmp_vars.box_selecting[0], xmax=event.mouse_region_x,
                                                  ymin=context.window_manager.f2f_tmp_vars.box_selecting[1], ymax=event.mouse_region_y, 
                                                  wait_for_input=False, mode='SUB') """
                        modal.report({'INFO'}, "BOX SELECTION SUBTRACTIVE:")#+str(context.window_manager.f2f_tmp_vars.box_selecting[0])+", "+str(context.window_manager.f2f_tmp_vars.box_selecting[0])+" :"+str(context.window_manager.f2f_tmp_vars.box_selecting[2])+", "+str(context.window_manager.f2f_tmp_vars.box_selecting[3]))
                    self.update_selection(context, modal) 
                    context.window_manager.f2f_tmp_vars.box_selecting = []
                    self.select_mode = 0
                    context.area.tag_redraw()
                    if not self.former_status_text is None:
                        modal.set_status_text(self.former_status_text, False)
                
                if event.value == 'PRESS':
                    context.window_manager.f2f_tmp_vars.box_selecting = [event.mouse_region_x, event.mouse_region_y, event.mouse_region_x, event.mouse_region_y]

            if event.type in {'ESC', 'RIGHTMOUSE'}:
                if event.value == 'RELEASE':
                    context.window_manager.f2f_tmp_vars.box_selecting = []
                    self.select_mode = 0
                    context.area.tag_redraw()
                    if not self.former_status_text is None:
                        modal.set_status_text(self.former_status_text, False)
                    modal.report({'INFO'}, "box selection left")

            if event.type == 'MOUSEMOVE' and len(context.window_manager.f2f_tmp_vars.box_selecting) > 0:
                context.area.tag_redraw()
                context.window_manager.f2f_tmp_vars.box_selecting = [context.window_manager.f2f_tmp_vars.box_selecting[0], context.window_manager.f2f_tmp_vars.box_selecting[1], event.mouse_region_x, event.mouse_region_y]

            self.mouse_start = (event.mouse_region_x,
                             event.mouse_region_y)
            return {'RUNNING_MODAL'}


        if self.select_mode == 5: #face loop select
            if context.window_manager.face2face.is_select(event.type):
                if event.value == 'RELEASE':
                    # select object on a loop ...
                    result = bpy.ops.view3d.select(extend=True, location=(event.mouse_region_x, event.mouse_region_y))
                    result = self.catch_wrong_selection(result, modal, context)
                    if not result is None:
                        return result

                    # if tf is not a quad there is no loop ... so do nothing 
                    if mesh.get_tft_type(context.active_object,modal.target_object) != 4:
                        # only set back to original selection and leave face loop select mode
                        if not context.active_object in modal.selected_objects:
                            context.view_layer.objects.active.select_set(False)
                        context.view_layer.objects.active = modal.projected_objects[modal.object_index]
                        self.select_mode = 0
                        modal.report({'INFO'}, "QUAD NEEDED AS TARGET TYPE")
                        context.view_layer.update()
                        context.area.tag_redraw()
                        context.window.cursor_set("DEFAULT")
                        if not self.former_status_text is None:
                            modal.set_status_text(self.former_status_text, False)
                        return {'RUNNING_MODAL'}
                    # self.update_selection(context, modal)

                    # ... determe its target face as startin point...
                    tf = context.active_object.target_face_index
                     
                    # ... to select a face loop on the target object
                    context.view_layer.objects.active = modal.target_object
                    modal.target_object.select_set(True)
                    bpy.ops.object.mode_set(mode='EDIT', toggle=False)
                    # with alt pressed we toggle the face loop direction
                    loop_faces = mesh.get_face_loop_faces(modal.target_object, tf, event.alt)
                    
                    # add or remove projected objects on this loop
                    if event.ctrl:
                        self.deselect_by_target_faces(context, modal, loop_faces)
                    else:
                        self.select_by_loop(context, modal, loop_faces, event.shift)

                    # deselect target object ...
                    modal.target_object.select_set(False)
                    context.view_layer.objects.active = modal.get_projected_object_by_tft(tf)
                    bpy.ops.object.mode_set(mode='OBJECT', toggle=False)

                    # and select the new determed selection ...
                    self.update_selection(context, modal) 
                    self.select_mode = 0
                    modal.report({'INFO'}, "FACE LOOP SELECTED")
                    context.area.tag_redraw()
                    if not self.former_status_text is None:
                        modal.set_status_text(self.former_status_text, False)
                    context.window.cursor_set("DEFAULT")
            self.mouse_start = (event.mouse_region_x,
                             event.mouse_region_y)
                    
            if event.type in {'ESC'} or context.window_manager.face2face.is_non_select(event.type):
                if event.value == 'RELEASE':
                    self.select_mode = 0
                    context.window.cursor_set("DEFAULT")
                    modal.report({'INFO'}, "face loop selection cancelled")
                    context.area.tag_redraw()
                    if not self.former_status_text is None:
                        modal.set_status_text(self.former_status_text, False)
                    self.mouse_start = (event.mouse_region_x,
                                event.mouse_region_y)

            return {'RUNNING_MODAL'}

        if self.select_mode == 0: # manually select object    
            if context.window_manager.face2face.is_select(event.type):# in context.window_manager.face2face.right_click_select:  #
                if event.value == 'RELEASE':            
                    if event.shift or event.alt:
                        return self.manual_selection(event.shift, event, modal, context)
        if self.select_mode == 1: # manually select object    
            if context.window_manager.face2face.is_select(event.type):# in context.window_manager.face2face.right_click_select:
                
                if event.value == 'RELEASE':
                   return self.manual_selection(event.shift, event, modal, context, self.former_status_text)
                   """  # result = bpy.ops.view3d.select(extend=True, location=(event.mouse_region_x, event.mouse_region_y))
                    result = bpy.ops.view3d.select( toggle=event.shift, location=(event.mouse_region_x, event.mouse_region_y))
                    context.view_layer.update()
                    result = self.catch_wrong_selection(result, modal, context)
                    if not result is None:
                        return result
                    
                    # bpy.ops.view3d.select( toggle=event.shift, location=(event.mouse_region_x, event.mouse_region_y))
                    context.view_layer.update()
                    self.update_selection(context, modal)
                    if not event.shift : #leave mode
                        self.select_mode = 0
                        modal.report({'INFO'}, "manualy selected one object")
                        self.mouse_start = (event.mouse_region_x,
                                        event.mouse_region_y)
                        context.window.cursor_set("DEFAULT")
                        modal.set_status_text(self.former_status_text, False)
                    else:
                        modal.report({'INFO'}, "manualy toggled an object")
                        context.area.tag_redraw()
                    return {'RUNNING_MODAL'} """

            if event.type in {'ESC'} or context.window_manager.face2face.is_non_select(event.type):
                if event.value == 'RELEASE':
                    self.select_mode = 0
                    modal.report({'INFO'}, "manualy selection cancelled")
                    context.area.tag_redraw()
                    if not self.former_status_text is None:
                        modal.set_status_text(self.former_status_text, False)
                    self.mouse_start = (event.mouse_region_x,
                                event.mouse_region_y)
            return {'RUNNING_MODAL'}

    def manual_selection(self, toggle, event, modal, context, former_status_text=None):
        result = bpy.ops.view3d.select(toggle=toggle, location=(event.mouse_region_x, event.mouse_region_y))
        context.view_layer.update()
        result = self.catch_wrong_selection(result, modal, context)
        if not result is None:
            return result
        
        # bpy.ops.view3d.select( toggle=event.shift, location=(event.mouse_region_x, event.mouse_region_y))
        context.view_layer.update()
        self.update_selection(context, modal)
        if not toggle : #leave mode if in M Mode
            self.select_mode = 0
            modal.report({'INFO'}, "manualy selected one object")
            self.mouse_start = (event.mouse_region_x,
                            event.mouse_region_y)
            context.window.cursor_set("DEFAULT")
            if not self.former_status_text is None:
                modal.set_status_text(self.former_status_text, False)
            context.area.tag_redraw()
        else:
            modal.report({'INFO'}, "manualy toggled an object")
            context.area.tag_redraw()
        return {'RUNNING_MODAL'}


    def catch_wrong_selection(self, result, modal, context):
        if 'FINISHED' not in result:
            self.select_mode = 0
            modal.report({'INFO'}, "NOTHING CHOOSEN")
            context.area.tag_redraw()
            context.window.cursor_set("DEFAULT")
            if not self.former_status_text is None:
                print("self.former_status_text:",self.former_status_text)
                modal.set_status_text(self.former_status_text, False)
            return {'RUNNING_MODAL'}

        projected_obj_selected = context.active_object in modal.projected_objects
        
        if not projected_obj_selected: # wrong selection do nothing
            # set back to original selection and leave face loop select mode
            self.update_selection(context, modal)
            # context.view_layer.objects.active = modal.projected_objects[modal.object_index]
            self.select_mode = 0
            modal.report({'INFO'}, "AN PROJECTED OBJECT NEEDS TO BE CHOOSEN")
            context.view_layer.update()
            context.area.tag_redraw()
            context.window.cursor_set("DEFAULT")
            if not self.former_status_text is None:
                modal.set_status_text(self.former_status_text, False)
            return {'RUNNING_MODAL'}

    def constrain_cursor(self, context, event):
        """ handle events for every state """      
        mouse_y = event.mouse_region_y + context.region.y
        mouse_x = event.mouse_region_x + context.region.x
        # mouse left 3d view to the right
        if mouse_x > (context.region.x + context.region.width):
            # set mouse back to left side     
            new_pos = (context.region.x + 5, mouse_y)
            context.window.cursor_warp(new_pos[0],new_pos[1])
            self.mouse_start = (new_pos[0]-context.region.x,new_pos[1]-context.region.y)
        # mouse left 3d view to the left
        if mouse_x < context.region.x + 5:
            # set mouse back to right side     
            new_pos = ((context.region.x + context.region.width), mouse_y)
            context.window.cursor_warp(new_pos[0],new_pos[1])
            self.mouse_start = (new_pos[0]-context.region.x,new_pos[1]-context.region.y)
        # mouse left 3d view upwards
        if mouse_y < context.region.y:
            # set mouse back to the bottom
            new_pos = (mouse_x, (context.region.y + context.region.height))          
            context.window.cursor_warp(new_pos[0],new_pos[1])
            self.mouse_start = (new_pos[0]-context.region.x,new_pos[1]-context.region.y)
        # mouse left 3d view downwards
        if mouse_y > (context.region.y + context.region.height):
            # set mouse back to the top     
            new_pos = (mouse_x, context.region.y)
            context.window.cursor_warp(new_pos[0],new_pos[1])
            self.mouse_start = (new_pos[0]-context.region.x,new_pos[1]-context.region.y)

    def handle_events(self, context, event, modal):
        """ handle events for every state """
        if event.type == "MOUSEMOVE":
            self.constrain_cursor(context, event)
        if event.type in {'UP_ARROW'}:
            if event.value == 'RELEASE':
                self.view_next(context, modal)
            return {'RUNNING_MODAL'}
        if event.type in {'DOWN_ARROW'}:
            if event.value == 'RELEASE':
                self.view_prev(context, modal)
            return {'RUNNING_MODAL'}

        if event.type in {'N'}:
            if event.value == 'RELEASE':
                modal.report({'INFO'}, "UI_TOGGLE")
                context.area.tag_redraw()
                context.space_data.show_region_ui = not context.space_data.show_region_ui
            return {'RUNNING_MODAL'}

        """ if event.type in {'M'}:
            if event.value == 'RELEASE':
                self.select_mode = 1
                context.window.cursor_set("EYEDROPPER")
                self.former_status_text = modal.status_text
                modal.set_status_text(" - MANUAL SELECT")
                modal.report({'INFO'}, "MANUAL SELECT MODE")
                context.area.tag_redraw()
            return {'RUNNING_MODAL'} """

        if event.type in {'T'}:
            if event.value == 'RELEASE':
                self.select_mode = 2
                context.window.cursor_set("EYEDROPPER")
                self.former_status_text = modal.status_text
                modal.set_status_text(" - OBJECT TYPE SELECTION")
                modal.report({'INFO'}, "OBJECT TYPE SELECTION MODE")
                context.area.tag_redraw()
            return {'RUNNING_MODAL'}

        if event.type in {'B'}:
            if event.value == 'RELEASE':
                self.former_status_text = modal.status_text
                modal.set_status_text(" - BOX SELECTION")
                self.select_mode = 4
                modal.set_status_text(" - BOX SELECTION")
                modal.report({'INFO'}, "BOX SELECTION MODE")
                # context.window.cursor_set("DOT")
                
            return {'RUNNING_MODAL'}

        if event.type in {'F'}:
            if event.value == 'RELEASE':
                self.select_mode = 5
                context.window.cursor_set("EYEDROPPER")
                self.former_status_text = modal.status_text
                modal.set_status_text(" - FACE LOOP SELECTION")
                modal.report({'INFO'}, "FACE LOOP SELECTION MODE")
                context.area.tag_redraw()
            return {'RUNNING_MODAL'}

        if event.type in {'C'}:
            if event.value == 'RELEASE':
                self.select_mode = 3
                context.window.cursor_set("DOT")
                context.window_manager.f2f_tmp_vars.circle_selecting = [event.mouse_region_x, event.mouse_region_y, 25]
                self.former_status_text = modal.status_text
                modal.set_status_text(" - CIRCLE SELECTION")
                modal.report({'INFO'}, "CIRCLE SELECTION MODE")
                context.area.tag_redraw()
            return {'RUNNING_MODAL'}



        # TODO solve problems with remapped hotkeys for navigation
        """ if event.type == 'MIDDLEMOUSE':
            if event.value == 'PRESS':
                self.view_button__pressed = True    
            return {'PASS_THROUGH'} """

        if keymap.is_user_relevant_key(event):#event.type == 'MIDDLEMOUSE':
            if event.value == 'PRESS':
                self.view_button_pressed = True    
            return {'PASS_THROUGH'}



        if event.type in {'WHEELUPMOUSE', 'WHEELDOWNMOUSE'}:
            return {'PASS_THROUGH'}
        # to prevent messing up the state
        # print("event.type :", event.type )
        return {'RUNNING_MODAL'}

    
    def reset_values4by_num(self):
        self.num_enter_mode = False
        self.num_value = 0
        self.decimal_place = 0
        self.period_used = False

    def check_num_input(self, context, event, modal):
        """ set the operation(grab,scale,rotate) by an entered concrete value  """
        if event.type in {"ONE", "TWO", "THREE", "FOUR", "FIVE", "SIX", "SEVEN", "EIGHT", "NINE", "ZERO", "MINUS", "PERIOD", "BACK_SPACE"}:
            if event.value == 'RELEASE':
                if event.alt or self.num_enter_mode or event.type == "PERIOD":
                    self.set_value_by_num(context, modal, event.type)
                    return {'RUNNING_MODAL'}

        if self.num_enter_mode:
            if event.type in {'RIGHTMOUSE', 'ESC'}:
                if event.value == 'RELEASE':
                    self.reset_values4by_num()
                    self.leave_num_Enter_mode(modal, context, False)
            if event.type in {'LEFTMOUSE', 'RET'}:
                if event.value == 'RELEASE':
                    self.reset_values4by_num()
                    self.leave_num_Enter_mode(modal, context, True)

            return {'RUNNING_MODAL'}

    def leave_num_Enter_mode(self, modal, context, apply):
        """ leave the precise key input mode  """
        if apply:
                modal.set_next_state("IDLE")
                return {'RUNNING_MODAL'}

        else:
                self.reset(context, modal)
                modal.set_next_state("IDLE")
                return {'RUNNING_MODAL'}

    def set_value_by_num(self, context, modal, num_string):
        """ set the operation(grab,scale,rotate) by an entered concrete value  """
        self.num_enter_mode = True

        if num_string == "MINUS":
            self.num_value *= -1
            return True

        if num_string == "BACK_SPACE":
            if self.period_used:
                if self.decimal_place == 0 or (self.decimal_place == 1 and self.num_value < 1):
                    self.reset_values4by_num()
                    
                    return False
                else:
                    self.decimal_place -=1
                    self.num_value = helpers.truncate(self.num_value, self.decimal_place)
                    
                    return True
            else:
                if self.num_value > 9:
                    self.num_value = math.floor(self.num_value/10)
                    return True
                else:
                    self.reset_values4by_num()
                    return False

        if num_string == "PERIOD":
            if self.period_used:
                return True
            else:
                self.period_used = True
                return True

        if self.period_used:
            num = self.map_string2num(num_string)
            self.decimal_place += 1
            self.num_value = self.num_value + num/(math.pow(10,self.decimal_place))
        else:
            num = self.map_string2num(num_string)
            self.num_value = self.num_value * 10 + num
        return True

    def enter_state(self, context, event, modal):
        """ just set some parameter at the beginning of the state """
        for obj in modal.projected_objects:
            obj.matrix_local_org = helpers.flatten(obj.matrix_local.copy())

    def leave_state(self, modal):
        """ leaving the state - cleanup  """
