""" class for the modal scaling state """
import math
import mathutils
import bpy

if "bpy" in locals():
    import importlib
    from .. import mesh
    importlib.reload(mesh)
    from .. import view
    importlib.reload(view)
    from . import base
    importlib.reload(base)
    from . base import ModalState
    from .. import helpers
    importlib.reload(helpers)
else:
    from .. import mesh
    from .. import view
    from . import base
    from . base import ModalState
    from .. import helpers


class ScaleState(ModalState):
    """ class for the modal scaling state """
    name = "Scaling"
    # mouse_start = (0, 0)
    axis_locked = 0
    limit_to_face = True
    inside = True
    fit_scale = False
    even_scale = True

    setting_variation = False
    setting_relative = False

    # obj_scaling_start = 1
    # the global variation at beginning of setting it
    variation_strength_start = 0
    relative_strength_start = 0

    # the global variation at entering the state
    variation_strength_start_state = 0
    relative_strength_start_state = 0

    def adapt_scale(self, modal, context):
        """ adopt scale from active to selected """
        active_obj = modal.projected_objects[modal.object_index]
        for obj in modal.selected_objects:
            obj.scaling = active_obj.scaling_org
            obj.scaling_vary = active_obj.scaling_vary
            scale_vec = mathutils.Vector((0, 0, 0))
            self.scale_on_faces_for(modal.selected_objects, context, 0, 0, modal, scale_vec)

    def set_axis(self, context, event, modal):
        if event.type in {'X'}:
            if event.value == 'RELEASE':
                if self.axis_locked in (0, 2, 3):
                    self.axis_locked = 1
                    modal.report({'INFO'}, "Scaling mode locked to x axis")
                else:
                    self.axis_locked = 0
                    modal.report({'INFO'}, "Scaling mode axis unlocked")
                if self.fit_scale:
                        self.reset2start(modal)
                        self.scale_fit_on_faces(context, modal)
                context.area.tag_redraw()
            return {'RUNNING_MODAL'}
        if event.type in {'Y'}:
            if event.value == 'RELEASE':
                if self.axis_locked in (0, 1, 3):
                    self.axis_locked = 2
                    modal.report({'INFO'}, "Scaling mode locked to y axis")
                else:
                    self.axis_locked = 0
                    modal.report({'INFO'}, "Scaling mode axis unlocked")
                if self.fit_scale:
                        self.reset2start(modal)
                        self.scale_fit_on_faces(context, modal)
                context.area.tag_redraw()
            return {'RUNNING_MODAL'}
        if event.type in {'Z'}:
            if event.value == 'RELEASE':
                if not self.fit_scale:
                    if self.axis_locked in (0, 1, 2):
                        self.axis_locked = 3
                        modal.report({'INFO'}, "Scaling mode locked to z axis")
                    else:
                        self.axis_locked = 0
                        modal.report({'INFO'}, "Scaling mode axis unlocked")
                context.area.tag_redraw()
            return {'RUNNING_MODAL'}


    def handle_events_vary(self, context, event, modal):
        """ handle events for the scaling state in set variation mode """
        if event.type == 'MOUSEMOVE':
            self.set_variation_strength(context, event.mouse_region_x,
                                    event.mouse_region_y, modal)
            self.scale_on_faces_for(modal.projected_objects, context, event.mouse_region_x,
                                    event.mouse_region_y, modal)
            # return {'RUNNING_MODAL'}

        if event.type in {'LEFTMOUSE', 'RET'}:
            if event.value == 'RELEASE':
                self.setting_variation = False
                modal.report({'INFO'}, "Variation strength set to:" +
                             str(modal.scale_variation_strength))
                context.area.tag_redraw()
                modal.set_status_text(" - SCALING")
                return {'RUNNING_MODAL'}

        if event.type in {'RIGHTMOUSE', 'ESC'}:
            if event.value == 'RELEASE':
                modal.scale_variation_strength = self.variation_strength_start
                self.setting_variation = False
                self.scale_on_faces_all(context, event.mouse_region_x,
                                        event.mouse_region_y, modal)
                modal.report({'INFO'}, "setting variation strength cancelled")
                context.area.tag_redraw()
                modal.set_status_text(" - SCALING")
                return {'RUNNING_MODAL'}

        if event.type in {'V'}:
            if event.value == 'RELEASE':
                self.set_per_object_variation(modal, context)
                self.scale_on_faces_all(context, event.mouse_region_x,
                                        event.mouse_region_y, modal)
                modal.report({'INFO'}, "New random scale variation for selected objects shuffled")

                return {'RUNNING_MODAL'}

        if event.type in {'Z'}:
            if event.value == 'RELEASE':
                self.zero_per_object_variation_for(modal.selected_objects)
                self.scale_on_faces_all(context, event.mouse_region_x,
                                        event.mouse_region_y, modal)
                modal.report(
                    {'INFO'}, "random scale variation set to zero for selected objects")
            return {'RUNNING_MODAL'}

    def handle_events_relative(self, context, event, modal):
        """ handle events for the scaling state in set relative mode """
        if event.type == 'MOUSEMOVE':
            self.set_relative_strength(context, event.mouse_region_x,
                                          event.mouse_region_y, modal)
            self.scale_on_faces_all(context, event.mouse_region_x,
                                        event.mouse_region_y, modal)
            # return {'RUNNING_MODAL'}

        if event.type in {'LEFTMOUSE', 'RET'}:
            if event.value == 'RELEASE':
                self.setting_relative = False
                modal.report({'INFO'}, "Face relation strength set to:" +
                             str(modal.scale_relative_strength))
                context.area.tag_redraw()
                return {'RUNNING_MODAL'}

        if event.type in {'R', 'RIGHTMOUSE', 'ESC'}:
            if event.value == 'RELEASE':
                modal.scale_relative_strength = self.relative_strength_start
                self.setting_relative = False
                self.scale_on_faces_all(context, event.mouse_region_x,
                                        event.mouse_region_y, modal)
                modal.report(
                    {'INFO'}, "setting face relation strength cancelled")
                context.area.tag_redraw()
                return {'RUNNING_MODAL'}

    def handle_events_normal(self, context, event, modal):
        """ handle events for the scaling state in normal mode """
        # print("event.type :", event.type )
        # print("event.value :", event.value )
        # return {'RUNNING_MODAL'}

        # if event.type == 'MIDDLEMOUSE':# or event.type == 'LEFTMOUSE' or event.value == 'RELEASE':
        #    print("event.type :", event.type )
        #    print("event.value :", event.value )
        #    self.mouse_start = (event.mouse_region_x,
        #                    event.mouse_region_y)
        # return {'RUNNING_MODAL'}

        if event.type == 'MOUSEMOVE':

            if self.view_button_pressed:
                self.mouse_start = (event.mouse_region_x,
                                event.mouse_region_y)
                self.view_button_pressed = False
            if not self.fit_scale:  # and not self.view_button_pressed:
                self.scale_on_faces_all(context, event.mouse_region_x,
                                    event.mouse_region_y, modal)
            # return {'RUNNING_MODAL'}

        if event.type in {'LEFTMOUSE', 'RET'}:
            if event.value == 'RELEASE':
                if self.fit_scale:
                    self.variation_strength_start = 0
                    self.relative_strength_start = 0
                    modal.scale_variation_strength = 0
                    modal.scale_relative_strength = 0
                if self.legal_apply():
                    modal.set_next_state("IDLE")
            return {'RUNNING_MODAL'}

        if event.type in {'RIGHTMOUSE', 'ESC'}:
            if event.value == 'RELEASE':
                self.reset(context, modal)
                modal.set_next_state("IDLE")
            return {'RUNNING_MODAL'}

        if event.type in {'V'}:
            if event.value == 'RELEASE':
                if not self.setting_variation and not self.fit_scale:
                    self.variation_strength_start = modal.scale_variation_strength
                    self.setting_variation = True
                    modal.set_status_text(" - SET RANDOM SCALE")
                    modal.report({'INFO'}, "Var set :"+str(modal.scale_variation_strength))
                    context.area.tag_redraw()

            return {'RUNNING_MODAL'}
    

        if event.type in {'R'}:
            if event.value == 'RELEASE':
                if not self.setting_relative and not self.fit_scale:
                    self.relative_strength_start = modal.scale_relative_strength
                    self.setting_relative = True
                    modal.report({'INFO'}, "Set random scale:")
                    context.area.tag_redraw()
            return {'RUNNING_MODAL'}


        if event.type in {'S'}:
            if event.value == 'RELEASE':
                if self.fit_scale:
                    # leave fit scale 
                    self.fit_scale = False
                    self.even_scale = False
                    self.mouse_start = (event.mouse_region_x,
                            event.mouse_region_y)
                    self.scale_on_faces_all(context, event.mouse_region_x,
                                    event.mouse_region_y, modal)
        """ if event.type in {'S'}:
            if event.value == 'RELEASE':
                self.fit_scale = not self.fit_scale
                if self.fit_scale:
                    self.reset2start(modal)
                    self.scale_fit_on_faces(context, modal)
                    modal.report({'INFO'}, "Scaling mode set to STRETCH FIT target faces")
                else:
                    self.mouse_start = (event.mouse_region_x,
                                event.mouse_region_y)
                    self.scale_on_faces_all(context, event.mouse_region_x,
                                    event.mouse_region_y, modal)
                    modal.report({'INFO'}, "Scaling mode set back to normal")
            return {'RUNNING_MODAL'}

        if event.type in {'E'}:
            if event.value == 'RELEASE':            
                if self.fit_scale:
                    if not self.even_scale:
                        self.even_scale = not self.even_scale
                        self.reset2start(modal)
                        self.scale_fit_on_faces(context, modal)
                        modal.report({'INFO'}, "Scaling mode STRETCH IT set to even")                    
            return {'RUNNING_MODAL'} """

        if event.type in {'E'}:
            if event.value == 'RELEASE':            
                if self.fit_scale:
                    if not self.even_scale:
                        # from i max back to even
                        self.even_scale = not self.even_scale
                        self.reset2start(modal)
                        self.scale_fit_on_faces(context, modal)
                        modal.report({'INFO'}, "Scaling mode STRETCH IT set to even")
                    else:
                        # leave fit scale 
                        self.even_scale = False
                        self.fit_scale = False
                        self.mouse_start = (event.mouse_region_x,
                                event.mouse_region_y)
                        self.scale_on_faces_all(context, event.mouse_region_x,
                                        event.mouse_region_y, modal)
                else:
                    # enter fit scale even
                    self.fit_scale = True
                    self.even_scale = True
                    self.reset2start(modal)
                    self.scale_fit_on_faces(context, modal)
                    modal.report({'INFO'}, "Scaling mode set to STRETCH FIT target faces")
            return {'RUNNING_MODAL'}

        if event.type in {'I'}:
            if event.value == 'RELEASE':
                if self.fit_scale:            
                    if self.even_scale:
                        # from even to i max
                        self.even_scale = not self.even_scale
                        self.reset2start(modal)
                        modal.report({'INFO'}, "Scaling mode STRETCH FIT set to individual max")
                        self.scale_fit_on_faces(context, modal)
                    else:
                        # leave fit scale 
                        self.fit_scale = False
                        self.even_scale = False
                        self.mouse_start = (event.mouse_region_x,
                                event.mouse_region_y)
                        self.scale_on_faces_all(context, event.mouse_region_x,
                                        event.mouse_region_y, modal)
                else:
                    # enter fit scale i max
                    self.even_scale = False
                    self.fit_scale = True
                    self.reset2start(modal)
                    self.scale_fit_on_faces(context, modal)
                    modal.report({'INFO'}, "Scaling mode set to STRETCH FIT target faces")
            return {'RUNNING_MODAL'}

        if event.type in {'L'}:
            if event.value == 'RELEASE':
                if not self.fit_scale:
                    self.limit_to_face = not self.limit_to_face
                    on_text = "ON" if self.limit_to_face else "OFF"
                    modal.report({'INFO'}, "Limit "+on_text)
                context.area.tag_redraw()
            return {'RUNNING_MODAL'}

        if event.type in {'A'}:
            if event.value == 'RELEASE':
                self.adapt_scale(modal, context)
            return {'RUNNING_MODAL'}

        if event.type in {'O'}:
            if event.value == 'RELEASE':
                self.reset2start(modal)
                if self.fit_scale:
                    self.fit_scale = False
            return {'RUNNING_MODAL'}

    def handle_events(self, context, event, modal):
        """ handle events for the scaling state """
        return_val = self.set_axis(context, event, modal)
        if return_val:
            if self.num_enter_mode: 
                self.scale_by_num_value(context, modal)
            return return_val
        return_val = self.check_num_input(context, event, modal)
        if return_val: return return_val
        # if self.select_mode > 0:
        return_val = self.handle_events_pre(context, event, modal)
        if return_val: return return_val

        if self.setting_variation:
            return_val = self.handle_events_vary(context, event, modal)
            if return_val: return return_val

        if self.setting_relative:
            return_val = self.handle_events_relative(context, event, modal)
            if return_val: return return_val

        if not self.setting_variation and not self.setting_relative:
            return_val = self.handle_events_normal(context, event, modal)
            if return_val: return return_val
 
        return super().handle_events(context, event, modal)

    def legal_apply(self):
        """ checks if the current state of the move manipulation could be applied """
        # in relative mode everything must be inside ... it is if manipulator is
        # if self.relative_pos:
        #    return self.inside
        
        # if no limit is set everything is legal  
        if not self.limit_to_face or self.fit_scale:
            return True
        # in absolute mode everything must be inside ... it is if manipulator is an all others
        # inside = self.inside
        # for
        
        # Caution HACK !!!
        # check is done in draw.highlight_target_faces to avoid double calculation
        return self.inside

    def set_target_face_type(self, context, modal, face_type_string):
        """ set the type of target face (tries, quads, etc) which were selected last?
            and automatically align view if wanted  """
        # determe next object index
        num = self.map_string2num(face_type_string)
        if num in modal.target_face_types:
            self.target_face_type = num
        if num == 1:
            self.target_face_type = num


    def set_value_by_num(self, context, modal, num_string):
        if not super().set_value_by_num(context, modal, num_string):
            return self.leave_num_Enter_mode(modal, context, False)    
        modal.set_status_text(" - SCALING - KEY INPUT")
        if self.num_value == 0:
            return

        self.num_value = helpers.clamp(self.num_value, 0.001, 1000)
        self.scale_by_num_value(context, modal)
        
    def scale_by_num_value(self, context, modal):
        scale_vec = mathutils.Vector((self.num_value-1, self.num_value-1, self.num_value-1))
        if self.axis_locked == 1:
            scale_vec.y = 0
            scale_vec.z = 0

        if self.axis_locked == 2:
            scale_vec.x = 0
            scale_vec.z = 0

        if self.axis_locked == 3:
            scale_vec.x = 0
            scale_vec.y = 0
            
        self.reset(context, modal)

        self.scale_on_faces_for(modal.selected_objects, context, 0, 0, modal, scale_vec)#rotate_on_faces(0, 0, modal, math.radians(self.num_value))
    

    def reset(self, context, modal):
        """ reset everything to its original state at the beginning of scaling """
        # for obj in context.selected_objects:
        for obj in modal.projected_objects:
            obj.matrix_local = mathutils.Matrix(obj.matrix_local_org)
            # set back the stored scaling per object
            vec_sca = obj.matrix_local_org.decompose()[2].copy()
            
            obj.scaling = obj.scaling_org

            obj.scaling_vary = obj.scaling_vary_org
        # set back the global scaling variation strength
        modal.scale_variation_strength = self.variation_strength_start_state
        modal.scale_relative_strength = self.relative_strength_start_state

        context.view_layer.update()


    def enter_state(self, context, event, modal):
        """ just set some parameter at the beginning of the scale process """
        super().enter_state(context, event, modal)
        for obj in modal.projected_objects:
            # obj.matrix_local_org = helpers.flatten(obj.matrix_local.copy())
            obj.scaling_center = mesh.get_midpoint_by_active(obj)

            obj.scaling_org = obj.scaling
            obj.scaling_vary_org = obj.scaling_vary

        self.mouse_start = (event.mouse_region_x,
                            event.mouse_region_y)

            

        self.variation_strength_start_state = modal.scale_variation_strength
        self.relative_strength_start_state = modal.scale_relative_strength

        modal.set_status_text(" - SCALING")

        self.limit_to_face = context.window_manager.face2face.start_with_limit
        

    def leave_state(self, modal):
        """ leaving the state cleanup  """

    
    def set_variation_strength(self, context, mouse_x, mouse_y, modal):
        """ scale all selected/projected objects on their individual target faces """   
        diff_x = mouse_x - self.mouse_start[0]
        scale = diff_x/100.0
        self.mouse_start = (mouse_x, mouse_y)
        modal.scale_variation_strength += scale
        # modal.scale_variation_strength = helpers.clamp(modal.scale_variation_strength, 0, 1)
        modal.scale_variation_strength = helpers.clamp(modal.scale_variation_strength, 0, context.window_manager.face2face.max_scale_variation_strength)
        context.area.tag_redraw()

    def set_relative_strength(self, context, mouse_x, mouse_y, modal):
        """ scale all selected/projected objects on their individual target faces """
        diff_x = mouse_x - self.mouse_start[0]
        scale = diff_x/100.0
        self.mouse_start = (mouse_x, mouse_y)
        modal.scale_relative_strength += scale
        # modal.scale_relative_strength = helpers.clamp(modal.scale_relative_strength, 0, 1)
        modal.scale_relative_strength = helpers.clamp(modal.scale_relative_strength, 0, context.window_manager.face2face.max_scale_relative_strength)
        context.area.tag_redraw()
            
    @staticmethod
    def set_per_object_variation(modal, context):
        """ shuffeling new random values per object """
        for obj in modal.projected_objects:
            obj.scaling_vary = helpers.get_random_variation()#random.randrange(1,100)/100
            if context.window_manager.face2face.vary_scale_step > 0:
                rest = obj.scaling_vary % context.window_manager.face2face.vary_scale_step
                obj.scaling_vary -= rest

    def set_per_object_variation_for(self, objects, context):
        """ shuffeling new random values per object """
        for obj in objects:
            obj.scaling_vary = helpers.get_random_variation()
            if context.window_manager.face2face.vary_scale_step > 0:
                rest = obj.scaling_vary % context.window_manager.face2face.vary_scale_step
                obj.scaling_vary -= rest
            

    def zero_per_object_variation_for(self, objects):
        """ shuffeling new random values per object """
        for obj in objects:
            obj.scaling_vary = 1

    def calc_scale(self, mouse_x, mouse_y):
        """ calculate the amount of scale by the mouse movement """
        diff_x = mouse_x - self.mouse_start[0]
        # diff_y = mouse_y - self.mouse_start[1]
        scale = diff_x/100.0
        # diff_vec = mathutils.Vector((diff_x, diff_y))
        # scale = diff_vec.length/100.0
        # if diff_x < diff_y and diff_x < 0: scale *= -1
        # if diff_y < diff_x and diff_y < 0: scale *= -1
        self.mouse_start = (mouse_x, mouse_y)

        scale_vec = mathutils.Vector((scale, scale, scale))
        if self.axis_locked == 1:
            scale_vec.y = 0
            scale_vec.z = 0

        if self.axis_locked == 2:
            scale_vec.x = 0
            scale_vec.z = 0

        if self.axis_locked == 3:
            scale_vec.x = 0
            scale_vec.y = 0

        return scale_vec

    def get_scale_vecs(self, obj,verts, target_object, tverts):
        """ calculate vecs to intersection points on target face """
        tverts_wc = mesh.converts_verts2wc(target_object, tverts)
        mid = mesh.get_midpoint_world(obj, verts)

        if self.axis_locked in(1,2):
            return self.get_scale_vecs_locked(obj, verts, target_object, tverts_wc, mid)

        verts_wc = mesh.converts_verts2wc(obj, verts)  

        i_point_vert_tuples = []
        for vert in verts_wc:
            i_point = mesh.get_intersection_point_polyedge_vec(tverts_wc, vert, mid, target_object, obj)
            if i_point is not None:
                i_point_vert_tuples.append((vert-mid, (i_point-mid)))

        return i_point_vert_tuples

    def get_scale_vecs_locked(self, obj, verts, target_object, tverts_wc, mid):
        """ calculate vecs to intersection points on target face on a locked axis """
        if self.axis_locked == 1:
            vec_x_wc = mathutils.Vector((100, 0, 0))
        if self.axis_locked == 2:
            vec_x_wc = mathutils.Vector((0, 100, 0))

        vec_x_local = mesh.convert_vert2wc(obj, vec_x_wc)
        i_point_vert_tuples = []

        for vert in verts:

            vert_wc = mesh.convert_vert2wc(obj, vert)
            vec_x_wc = vec_x_local+(vert_wc-obj.location)
            i_point = mesh.get_intersection_point_polyedge_vec(tverts_wc, vec_x_wc, vert_wc, target_object, obj)
            if i_point is not None:
                i_point_vert_tuples.append((vert-mesh.get_midpoint(verts), (mesh.convert_vert_wc2local(obj, i_point)-mesh.get_midpoint(verts))))

        return i_point_vert_tuples

    def get_shortest_vec(self, i_point_vert_tuples):
        """ calculate the shortest vec to the closest intersection point
            on the target face """
        shortest = None
        for iv_tuple in i_point_vert_tuples:
            vec = iv_tuple[0] - iv_tuple[1]
            if shortest is None:
                shortest = iv_tuple
            else:
                shortest_vec = shortest[0] - shortest[1]
                if vec.length_squared < shortest_vec.length_squared:
                    shortest = iv_tuple
                
        return shortest

    def calc_scale_fit(self, obj, modal):
        """ calculate the amount of scale needed to fit the projected objects
            on their maximum extends to their individual target faces """
        verts = mesh.get_verts_from_polygon(obj, mesh.get_active_polygon(obj))
        target_face = modal.target_object.data.polygons[obj.target_face_index]
        tverts = mesh.get_verts_from_polygon(modal.target_object, target_face)

        if not mesh.geo_inside_poly(tverts, modal.target_object, verts, obj):
            return mathutils.Vector((1.0, 1.0, 1.0))

        scale_vecs = self.get_scale_vecs(obj, verts, modal.target_object, tverts)
        scale_vec_tuple = self.get_shortest_vec(scale_vecs)

        if scale_vec_tuple is not None:

            scale = scale_vec_tuple[1].length/scale_vec_tuple[0].length
            scale_vec = mathutils.Vector((scale, scale, scale))

            if self.axis_locked == 1:
                scale_vec = mathutils.Vector((scale_vec_tuple[1].x/scale_vec_tuple[0].x, 1,1))

            if self.axis_locked == 2:
                scale_vec = mathutils.Vector((1, scale_vec_tuple[1].y/scale_vec_tuple[0].y,1))
        else:
            scale_vec = mathutils.Vector((1, 1, 1))

        return scale_vec

    def calc_scale_fit_even(self, modal):
        """ calculate the amount of scale needed to fit the projected objects 
            on their maximum extends to their individual target faces while keeping
            all scales of each object even by taking the minimum of their maximums"""
        scale_vecs = []
        for obj in modal.projected_objects:
            verts = mesh.get_verts_from_polygon(obj, mesh.get_active_polygon(obj))
            target_face = modal.target_object.data.polygons[obj.target_face_index]
            tverts = mesh.get_verts_from_polygon(modal.target_object, target_face)

            if not mesh.geo_inside_poly(tverts, modal.target_object, verts, obj):
                return mathutils.Vector((1.0, 1.0, 1.0))

            scale_vecs += self.get_scale_vecs(obj, verts, modal.target_object, tverts)

        scale_vec_tuple = self.get_shortest_vec(scale_vecs)

        if scale_vec_tuple is not None:

            scale = scale_vec_tuple[1].length/scale_vec_tuple[0].length
            scale_vec = mathutils.Vector((scale, scale, scale))

            if self.axis_locked == 1:
                scale_vec = mathutils.Vector((scale_vec_tuple[1].x/scale_vec_tuple[0].x, 1,1))

            if self.axis_locked == 2:
                scale_vec = mathutils.Vector((1, scale_vec_tuple[1].y/scale_vec_tuple[0].y,1))
        else:
            scale_vec = mathutils.Vector((1, 1, 1))     

        return scale_vec


    def calc_rel_size(self, obj, modal):
        """ calculate the relative size of an objects target face to the current one """
        tpoly = modal.target_object.data.polygons[obj.target_face_index]
        size = tpoly.area
        return size/self.calc_median_size(modal)#TODO calc median only once and store in modal

    def calc_median_size(self, modal):
        """ calculate the median size of all projected objects """
        median_size = 0
        for obj in modal.projected_objects:
            current_tpoly = modal.target_object.data.polygons[obj.target_face_index]
            median_size += current_tpoly.area

        return median_size/len(modal.projected_objects)


    def scale_on_faces_all(self, context, mouse_x, mouse_y, modal):
        """ scale all selected/projected objects on their individual target faces """     
        self.scale_on_faces_for(modal.selected_objects, context, mouse_x, mouse_y, modal)


    MIN_SCALE = 0.001
    def set_scale_by_vec(self, obj, scale_vec):
        """ set scaling by scale vec and do not get under MIN_SCALE """
        obj.scaling = mathutils.Vector((obj.scaling[0]*(1+scale_vec.x), obj.scaling[1]*(1+scale_vec.y), obj.scaling[2]*(1+scale_vec.z)))#  * (mathutils.Vector((1,1,1)) + scale_vec)
        
        if obj.scaling[0] < self.MIN_SCALE:
            obj.scaling[0] = self.MIN_SCALE
        if obj.scaling[1] < self.MIN_SCALE:
            obj.scaling[1] = self.MIN_SCALE
        if obj.scaling[2] < self.MIN_SCALE:
            obj.scaling[2] = self.MIN_SCALE

    def scale_on_faces_for(self, objects, context, mouse_x, mouse_y, modal, scale_vec=None):
        """ scale all selected/projected objects on their individual target faces """

        if scale_vec is None:
            scale_vec = self.calc_scale(mouse_x, mouse_y)
        for obj in objects:

            # obj.scaling = mathutils.Vector(obj.scaling)  * (mathutils.Vector((1,1,1)) + scale_vec)#+ scale_vec
            self.set_scale_by_vec(obj, scale_vec)

            if modal.scale_relative_strength > 0:
                rel_size = self.calc_rel_size(obj, modal)
                rel_size += (1-rel_size)*(1-modal.scale_relative_strength)
                scale_vec_obj = mathutils.Vector(obj.scaling) * rel_size
            else:
                scale_vec_obj = mathutils.Vector(obj.scaling) 

            # add variation to objects scale
            if modal.scale_variation_strength > 0:
                scale_vec_obj *= 1 + modal.scale_variation_strength * (obj.scaling_vary-0.5)#(obj.scaling_vary-1)
        
            mat_sca = mesh.get_sca_matrix(scale_vec_obj)

            # get location and rotation out of the obj transform matrix
            deco = obj.matrix_local.decompose()  
            mat_loc = mathutils.Matrix.Translation(deco[0])
            mat_rot = deco[1].to_matrix().to_4x4()

            # store mid of projection face, object pos and local matrix bevor altering
            org_mid = mesh.get_midpoint_by_active(obj)
            matrix_local_bv = obj.matrix_local.copy()
  
            # create new transform matrix from original scale and current loc, rotation
            obj.matrix_local = mat_loc.copy() @ mat_rot @ mat_sca

            # now projected face mid has changed ... fix that -->
            # transform offset to local mid into its original state
            matrix_local_bv.invert()
            local_mid = matrix_local_bv @ org_mid

            # transform original local mid back into altered current state
            new_mid = mat_loc.copy() @ mat_rot @ mat_sca @ local_mid

            # calculate the offset of the object through the scaling and put it back 
            loc_diff = org_mid - new_mid
            obj.location += loc_diff


    def scale_fit_on_faces(self, context, modal):
        """ scale all selected/projected objects on their individual target faces """
        if self.even_scale:
            scale_vec = self.calc_scale_fit_even(modal)

        for obj in modal.selected_objects:
            obj.select_set(False)

        for obj in modal.selected_objects:

            if not self.even_scale:
                
                scale_vec = self.calc_scale_fit(obj, modal)
                print("not even scale",scale_vec)
            # ...select only the current one to scale them one by one
            obj.select_set(True)

            org_local_rot = obj.matrix_local_startorg.decompose()[1].to_matrix()

            bpy.ops.transform.resize(
                value=(scale_vec.x, scale_vec.y, scale_vec.z),
                orient_matrix=obj.matrix_local.to_quaternion().to_matrix() ,
                center_override=(obj.scaling_center))
            obj.select_set(False)  # ... and deselect afterwards

            # update the stored scaling per object
            bpy.context.view_layer.update()
            obj_sca = obj.matrix_local.decompose()[2]
            obj.scaling = obj_sca

        # ...reselect the the formely deselected
        for obj in modal.selected_objects:
            obj.select_set(True)

    def reset2start(self, modal):
        """ reset scaling of all projected objects to its original state at the beginning """
        
        for obj in modal.projected_objects:
            # get scaling matrix from original start transform matrix
            vec_sca = obj.matrix_local_startorg.decompose()[2]
            mat_sca = mesh.get_sca_matrix(vec_sca)

            # get location and rotation out of the current transform matrix
            deco = obj.matrix_local.decompose()  
            mat_loc = mathutils.Matrix.Translation(deco[0])
            mat_rot = deco[1].to_matrix().to_4x4()

            # store mid of projection face bevor reseting     
            org_mid = mesh.get_midpoint_by_active(obj)
            
            # create new transform matrix from original scale and current loc,rot
            obj.matrix_local = mat_loc.copy() @ mat_rot @ mat_sca

            # the scaling sets the object off because it is not scaled around origin
            # ...so we need to set it back into its place
            bpy.context.view_layer.update()
            new_mid = mesh.get_midpoint_by_active(obj)
            loc_diff = org_mid - new_mid
            obj.location += loc_diff
            bpy.context.view_layer.update()

            # update the stored scaling per object
            obj.scaling = obj.scaling_startorg
        
        modal.scale_variation_strength = 0
        modal.scale_relative_strength = 0

